#  Copyright (c) 2016 IBM Corporation
#  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import eventlet
import six

from oslo_log import log as logging
from oslo_utils import excutils

from cinder import coordination
from cinder import exception
from cinder.i18n import _
import cinder.volume.drivers.ibm.ibm_storage as storage
from cinder.volume.drivers.ibm.ibm_storage import ds8k_helper as helper
from cinder.volume.drivers.ibm.ibm_storage import ds8k_restclient as restclient
from cinder.volume.drivers.ibm.ibm_storage import proxy

LOG = logging.getLogger(__name__)

PPRC_PATH_NOT_EXIST = 0x00
PPRC_PATH_HEALTHY = 0x01
PPRC_PATH_UNHEALTHY = 0x02
PPRC_PATH_FULL = 0x03


class MetroMirrorManager(object):
    """Manage metro mirror for replication."""

    def __init__(self, source_helper, target_helper):
        self._source_helper = source_helper
        self._target_helper = target_helper

    def switch_source_and_target(self):
        self._source_helper, self._target_helper = (
            self._target_helper, self._source_helper)

    def check_physical_links(self):
        ports = self._source_helper.get_physical_links(
            self._target_helper.backend['storage_wwnn'])
        if not ports:
            raise exception.VolumeDriverException(
                message=((_("%(tgt)s is not connected to %(src)s!") % {
                    'tgt': self._target_helper.backend['storage_wwnn'],
                    'src': self._source_helper.backend['storage_wwnn']
                })))

        pairs = [{
            'source_port_id': p['source_port_id'],
            'target_port_id': p['target_port_id']
        } for p in ports]
        if not self._target_helper.backend['port_pairs']:
            # if there are more than eight physical links,
            # choose eight of them.
            self._target_helper.backend['port_pairs'] = (
                pairs[:8] if len(pairs) > 8 else pairs)
        else:
            # verify the port pairs user set
            for pair in self._target_helper.backend['port_pairs']:
                if pair not in pairs:
                    valid_pairs = ';'.join(
                        ["%s-%s" % (p['source_port_id'],
                                    p['target_port_id'])
                         for p in pairs])
                    invalid_pair = "%s-%s" % (pair['source_port_id'],
                                              pair['target_port_id'])
                    raise exception.VolumeDriverException(
                        message=((_("Invalid port pair: %(invalid)s, valid "
                                    "port pair(s) are: %(valid)s")
                                  % {'invalid': invalid_pair,
                                     'valid': valid_pairs})))
        self._source_helper.backend['port_pairs'] = [{
            'source_port_id': p['target_port_id'],
            'target_port_id': p['source_port_id']
        } for p in self._target_helper.backend['port_pairs']]

    def is_target_alive(self):
        try:
            self._target_helper.get_systems()
        except restclient.TimeoutException as e:
            LOG.info("REST request time out, backend may be not available "
                     "any more. Exception: %s", e)
            return False

        return True

    def find_from_pprc_paths(self, specified_lss=None, excluded_lss=None):
        """find lss from existing pprc paths and pool id for it.

        the format of pool_lss_pair returned is as below:
        {'source': (pid, lss), 'target': (pid, lss)}
        """
        state, paths = self._filter_pprc_paths(specified_lss)
        if state != PPRC_PATH_HEALTHY:
            # check whether the physical links are available or not,
            # or have been changed.
            self.check_physical_links()
            return state, None
        if excluded_lss:
            paths = [p for p in paths
                     if p['source_lss_id'] not in excluded_lss]
        # only establish_replication will specify the source LSS
        # and it need to reuse LSS reserved for CG if this LSS
        # is in PPRC path.
        if not specified_lss:
            paths = [p for p in paths if p['source_lss_id'] not in
                     self._source_helper.backend['lss_ids_for_cg']]

        # sort pairs according to the number of luns in their LSSes,
        # and get the pair which LSS has least luns.
        candidates = []
        source_lss_set = set(p['source_lss_id'] for p in paths)
        for lss in source_lss_set:
            # get the number of luns in source.
            src_luns = self._source_helper.get_lun_number_in_lss(lss)
            if src_luns == helper.LSS_VOL_SLOTS and not specified_lss:
                continue

            spec_paths = [p for p in paths if p['source_lss_id'] == lss]
            for path in spec_paths:
                # get the number of luns in target.
                try:
                    tgt_luns = self._target_helper.get_lun_number_in_lss(
                        path['target_lss_id'])
                except restclient.APIException:
                    # if DS8K can fix this problem, then remove the
                    # exception here.
                    LOG.error("Target LSS %s in PPRC path may doesn't "
                              "exist although PPRC path is available.",
                              path['target_lss_id'])
                    tgt_luns = 0
                candidates.append((path['source_lss_id'],
                                   path['target_lss_id'],
                                   src_luns + tgt_luns))
        if not candidates:
            return PPRC_PATH_FULL, None
        else:
            src_lss, tgt_lss, num = sorted(candidates, key=lambda c: c[2])[0]
            return PPRC_PATH_HEALTHY, {
                'source': (self._source_helper.get_pool(src_lss), src_lss),
                'target': (self._target_helper.get_pool(tgt_lss), tgt_lss)
            }

    def _filter_pprc_paths(self, lss):
        paths = self._source_helper.get_pprc_paths(lss)
        if paths:
            # get the paths only connected to replication target
            paths = [p for p in paths if p['target_system_wwnn'] in
                     self._target_helper.backend['storage_wwnn']]
        else:
            LOG.info("No PPRC paths found in primary DS8K.")
            return PPRC_PATH_NOT_EXIST, None

        # get the paths whose port pairs have been set in configuration file.
        expected_port_pairs = [
            (port['source_port_id'], port['target_port_id'])
            for port in self._target_helper.backend['port_pairs']]
        for path in paths[:]:
            port_pairs = [(p['source_port_id'], p['target_port_id'])
                          for p in path['port_pairs']]
            if not (set(port_pairs) & set(expected_port_pairs)):
                paths.remove(path)
        if not paths:
            LOG.info("Existing PPRC paths do not use port pairs that "
                     "are set.")
            return PPRC_PATH_NOT_EXIST, None

        # abandon PPRC paths according to volume type(fb/ckd)
        source_lss_set = set(p['source_lss_id'] for p in paths)
        if self._source_helper.backend.get('device_mapping'):
            source_lss_set = source_lss_set & set(
                self._source_helper.backend['device_mapping'].keys())
        else:
            all_lss = self._source_helper.get_all_lss(['id', 'type'])
            fb_lss = set(
                lss['id'] for lss in all_lss if lss['type'] == 'fb')
            source_lss_set = source_lss_set & fb_lss
        paths = [p for p in paths if p['source_lss_id'] in source_lss_set]
        if not paths:
            LOG.info("No source LSS in PPRC paths has correct volume type.")
            return PPRC_PATH_NOT_EXIST, None

        # if the group property of lss doesn't match pool node,
        # abandon these paths.
        discarded_src_lss = []
        discarded_tgt_lss = []
        for lss in source_lss_set:
            spec_paths = [p for p in paths if p['source_lss_id'] == lss]
            if self._source_helper.get_pool(lss) is None:
                discarded_src_lss.append(lss)
                continue

            for spec_path in spec_paths:
                tgt_lss = spec_path['target_lss_id']
                if self._target_helper.get_pool(tgt_lss) is None:
                    discarded_tgt_lss.append(tgt_lss)

        if discarded_src_lss:
            paths = [p for p in paths if p['source_lss_id'] not in
                     discarded_src_lss]
        if discarded_tgt_lss:
            paths = [p for p in paths if p['target_lss_id'] not in
                     discarded_tgt_lss]
        if not paths:
            LOG.info("No PPRC paths can be re-used.")
            return PPRC_PATH_NOT_EXIST, None

        # abandon unhealthy PPRC paths.
        for path in paths[:]:
            failed_port_pairs = [
                p for p in path['port_pairs'] if p['state'] != 'success']
            if len(failed_port_pairs) == len(path['port_pairs']):
                paths.remove(path)
        if not paths:
            LOG.info("PPRC paths between primary and target DS8K "
                     "are unhealthy.")
            return PPRC_PATH_UNHEALTHY, None

        return PPRC_PATH_HEALTHY, paths

    def create_pprc_path(self, lun, is_group=False):
        switch = lun.failed_over if is_group else False
        src_helper, tgt_helper = (
            (self._target_helper, self._source_helper) if switch else
            (self._source_helper, self._target_helper))
        src_lss = lun.pool_lss_pair['source'][1]
        tgt_lss = lun.pool_lss_pair['target'][1]
        # check whether the pprc path exists and is healthy or not.
        pid = (src_helper.backend['storage_wwnn'] + '_' + src_lss + ':' +
               tgt_helper.backend['storage_wwnn'] + '_' + tgt_lss)
        state = self._is_pprc_paths_healthy(pid, switch)
        LOG.info("The state of PPRC path %(path)s is %(state)s.",
                 {'path': pid, 'state': state})
        if state == PPRC_PATH_HEALTHY:
            return

        # create the pprc path
        pathData = {
            'target_system_wwnn': tgt_helper.backend['storage_wwnn'],
            'source_lss_id': src_lss,
            'target_lss_id': tgt_lss,
            'port_pairs': tgt_helper.backend['port_pairs']
        }
        if lun.group and lun.group.consisgroup_replication_enabled:
            pathData['pprc_consistency_group'] = 'enable'
        LOG.info("PPRC path %(src)s:%(tgt)s will be created.",
                 {'src': src_lss, 'tgt': tgt_lss})
        src_helper.create_pprc_path(pathData)

        # check the state of the pprc path
        LOG.debug("Checking the state of the new PPRC path.")
        for retry in range(4):
            eventlet.sleep(2)
            if self._is_pprc_paths_healthy(pid, switch) == PPRC_PATH_HEALTHY:
                break
            if retry == 3:
                src_helper.delete_pprc_path(pid)
                raise restclient.APIException(
                    data=(_("Failed to create PPRC path %(src)s:%(tgt)s.")
                          % {'src': src_lss, 'tgt': tgt_lss}))
        LOG.debug("Create the new PPRC path successfully.")

    def _is_pprc_paths_healthy(self, path_id, switch):
        bck_helper = self._target_helper if switch else self._source_helper
        try:
            path = bck_helper.get_pprc_path(path_id)
        except restclient.APIException:
            return PPRC_PATH_NOT_EXIST

        for port in path['port_pairs']:
            if port['state'] == 'success':
                return PPRC_PATH_HEALTHY

        return PPRC_PATH_UNHEALTHY

    def create_pprc_pairs(self, lun):
        tgt_vol_id = lun.replication_driver_data[
            self._target_helper.backend['id']]['vol_hex_id']
        tgt_stg_id = self._target_helper.backend['storage_unit']

        vol_pairs = [{
            'source_volume': lun.ds_id,
            'source_system_id': self._source_helper.backend['storage_unit'],
            'target_volume': tgt_vol_id,
            'target_system_id': tgt_stg_id
        }]
        pair_data = {
            "volume_pairs": vol_pairs,
            "type": "metro_mirror",
            "options": ["permit_space_efficient_target",
                        "initial_copy_full"]
        }
        LOG.debug("Creating pprc pair, pair_data is %s.", pair_data)
        self._source_helper.create_pprc_pair(pair_data)
        self._source_helper.wait_pprc_copy_finished([lun.ds_id], 'full_duplex')
        LOG.info("The state of PPRC pair has become full_duplex.")

    def delete_pprc_pairs(self, lun):
        self._source_helper.delete_pprc_pair(lun.ds_id)
        if self.is_target_alive() and lun.replication_driver_data:
            replica = sorted(lun.replication_driver_data.values())[0]
            self._target_helper.delete_pprc_pair(replica['vol_hex_id'])

    def do_pprc_failover(self, luns, is_group=False):
        switch = luns[0].failed_over if is_group else False
        src_helper, tgt_helper = (
            (self._target_helper, self._source_helper) if switch else
            (self._source_helper, self._target_helper))
        vol_pairs = []
        target_vol_ids = []
        for lun in luns:
            if not tgt_helper.lun_exists(lun.replica_ds_id):
                LOG.info("Target volume %(volid)s doesn't exist in "
                         "DS8K %(storage)s.",
                         {'volid': lun.replica_ds_id,
                          'storage': tgt_helper.backend['storage_unit']})
                continue

            vol_pairs.append({
                'source_volume': lun.replica_ds_id,
                'source_system_id': tgt_helper.backend['storage_unit'],
                'target_volume': lun.ds_id,
                'target_system_id': src_helper.backend['storage_unit']
            })
            target_vol_ids.append(lun.replica_ds_id)

        pair_data = {
            "volume_pairs": vol_pairs,
            "type": "metro_mirror",
            "options": ["failover"]
        }

        LOG.info("Begin to fail over to %(backend)s, "
                 "pair_data is %(pair_data)s.",
                 {'backend': tgt_helper.backend['storage_unit'],
                  'pair_data': pair_data})
        tgt_helper.create_pprc_pair(pair_data)
        tgt_helper.wait_pprc_copy_finished(target_vol_ids,
                                           'suspended', switch)
        LOG.info("Failover from %(src)s to %(tgt)s is finished.", {
            'src': src_helper.backend['storage_unit'],
            'tgt': tgt_helper.backend['storage_unit']
        })

    def get_pprc_pair_ids(self, luns, switch=False):
        if not luns:
            return None
        src_helper, tgt_helper = (
            (self._target_helper, self._source_helper) if switch else
            (self._source_helper, self._target_helper))
        pprc_pair_ids = []
        for lun in luns:
            if switch:
                is_lun_exist = tgt_helper.lun_exists(lun.replica_ds_id)
            else:
                is_lun_exist = src_helper.lun_exists(lun.ds_id)
            if not is_lun_exist:
                LOG.info("Target volume %(volume)s doesn't exist in "
                         "DS8K %(storage)s.",
                         {'volume': (lun.replica_ds_id
                                     if switch else lun.ds_id),
                          'storage': (tgt_helper.backend['storage_unit']
                                      if switch else
                                      src_helper.backend['storage_unit'])})
                continue
            pprc_pair_ids.append(
                src_helper.backend['storage_unit'] + '_' + lun.ds_id + ':' +
                tgt_helper.backend['storage_unit'] + '_' + lun.replica_ds_id)
        return pprc_pair_ids

    def do_pprc_failback(self, luns, is_group=False):
        switch = luns[0].failed_over if is_group else False
        bck_helper = self._target_helper if switch else self._source_helper
        pair_data = {"pprc_ids": self.get_pprc_pair_ids(luns, switch),
                     "type": "metro_mirror",
                     "options": ["failback"]}
        LOG.info("Begin to run failback in %(backend)s, "
                 "pair_data is %(pair_data)s.",
                 {'backend': bck_helper.backend['storage_unit'],
                  'pair_data': pair_data})
        bck_helper.do_failback(pair_data)
        lun_ids = [lun.ds_id for lun in luns]
        bck_helper.wait_pprc_copy_finished(lun_ids, 'full_duplex', switch)
        LOG.info("Run failback in %s is finished.",
                 bck_helper.backend['storage_unit'])


class Replication(object):
    """Metro Mirror and Global Mirror will be used by it.

    Version history:

    .. code-block:: none

        1.0.0 - initial revision.
        2.1.0 - ignore exception during cleanup when creating or deleting
                replica failed.
        2.1.1 - Adding support for replication consistency group.
    """

    VERSION = "2.1.1"

    def __init__(self, source_helper, target_device):
        self._source_helper = source_helper
        connection_type = target_device.get('connection_type')
        if connection_type == storage.XIV_CONNECTION_TYPE_FC:
            self._target_helper = (
                helper.DS8KReplicationTargetHelper(target_device))
        elif connection_type == storage.XIV_CONNECTION_TYPE_FC_ECKD:
            self._target_helper = (
                helper.DS8KReplicationTargetECKDHelper(target_device))
        else:
            raise exception.InvalidParameterValue(
                err=(_("Param [connection_type] %s in replication_device "
                       "is invalid.") % connection_type))

        if self._target_helper.backend['lss_ids_for_cg']:
            if (len(self._target_helper.backend['lss_ids_for_cg']) !=
               len(self._source_helper.backend['lss_ids_for_cg'])):
                raise exception.VolumeDriverException(
                    message=_("Please reserve the same number of LSS for "
                              "secondary DS8K just as the primary DS8K."))
        else:
            self._target_helper.backend['lss_ids_for_cg'] = (
                self._source_helper.backend['lss_ids_for_cg'])

        self._mm_manager = MetroMirrorManager(self._source_helper,
                                              self._target_helper)

    def get_target_helper(self):
        return self._target_helper

    def get_source_helper(self):
        return self._source_helper

    def check_connection_type(self):
        src_conn_type = self._source_helper.get_connection_type()
        tgt_conn_type = self._target_helper.get_connection_type()
        if src_conn_type != tgt_conn_type:
            raise exception.VolumeDriverException(
                message=(_("The connection type in primary backend is "
                           "%(primary)s, but in secondary backend it is "
                           "%(secondary)s")
                         % {'primary': src_conn_type,
                            'secondary': tgt_conn_type}))

    def check_physical_links(self):
        self._mm_manager.check_physical_links()

    def switch_source_and_target_client(self):
        # switch the helper in metro mirror manager
        self._mm_manager.switch_source_and_target()
        # switch the helper
        self._source_helper, self._target_helper = (
            self._target_helper, self._source_helper)

    def _switch_source_and_target_volume(self, luns, secondary_backend_id):
        for lun in luns:
            if secondary_backend_id == 'default':
                backend_id = self._target_helper.backend['id']
                lun.failed_over = False
            else:
                backend_id = 'default'
                lun.failed_over = True
            # secondary_id is never blank here.
            lun.replication_driver_data = (
                {backend_id: {'vol_hex_id': lun.ds_id}})
            lun.ds_id, lun.replica_ds_id = lun.replica_ds_id, lun.ds_id
        return luns

    @proxy.logger
    def find_pool_lss_pair(self, excluded_lss):
        state, pool_lss_pair = (
            self._mm_manager.find_from_pprc_paths(None, excluded_lss))
        if pool_lss_pair is None:
            pool_lss_pair = self.find_new_lss_for_source(excluded_lss)
            pool_lss_pair.update(self.find_new_lss_for_target())
        return pool_lss_pair

    @proxy.logger
    def find_new_lss_for_source(self, excluded_lss):
        src_pid, src_lss = self._source_helper.find_pool_and_lss(excluded_lss)
        return {'source': (src_pid, src_lss)}

    @proxy.logger
    def find_new_lss_for_target(self):
        tgt_pid, tgt_lss = self._target_helper.find_pool_and_lss()
        return {'target': (tgt_pid, tgt_lss)}

    @proxy.logger
    def establish_replication(self, lun, delete_source=False):
        state, lun.pool_lss_pair = (
            self._mm_manager.find_from_pprc_paths(lun.ds_id[0:2]))
        LOG.debug("establish_replication: pool_lss_pair is %s.",
                  lun.pool_lss_pair)
        if state == PPRC_PATH_UNHEALTHY:
            raise restclient.APIException(
                data=(_("The path(s) for volume %(name)s isn't available "
                        "any more, please make sure the state of the path(s) "
                        "which source LSS is %(lss)s is success.")
                      % {'name': lun.cinder_name, 'lss': lun.ds_id[0:2]}))
        elif state == PPRC_PATH_NOT_EXIST:
            pid = self._source_helper.get_pool(lun.ds_id[0:2])
            lun.pool_lss_pair = {'source': (pid, lun.ds_id[0:2])}
            lun.pool_lss_pair.update(self.find_new_lss_for_target())
        lun = self.create_replica(lun, delete_source)
        return lun

    @proxy.logger
    @coordination.synchronized('ibm-ds8k-replication')
    def create_replica(self, lun, delete_source=True):
        try:
            self._target_helper.create_lun(lun)
            # create PPRC paths if need.
            self._mm_manager.create_pprc_path(lun)
            # create pprc pair
            self._mm_manager.create_pprc_pairs(lun)
        except restclient.APIException:
            with excutils.save_and_reraise_exception():
                try:
                    self.delete_replica(lun)
                    if delete_source:
                        self._source_helper.delete_lun(lun)
                except restclient.APIException as ex:
                    LOG.info("Failed to cleanup replicated volume %(id)s, "
                             "Exception: %(ex)s.",
                             {'id': lun.ds_id, 'ex': ex})
        lun.replication_status = 'enabled'
        return lun

    @proxy.logger
    def delete_replica(self, lun, delete_source=False):
        if lun.ds_id is not None:
            try:
                self._mm_manager.delete_pprc_pairs(lun)
                self._delete_replica(lun)
            except restclient.APIException as e:
                if delete_source:
                    try:
                        self._source_helper.delete_lun(lun)
                    except restclient.APIException as ex:
                        LOG.info("Failed to delete source volume %(id)s, "
                                 "Exception: %(ex)s.",
                                 {'id': lun.ds_id, 'ex': ex})
                raise exception.VolumeDriverException(
                    message=(_('Failed to delete the target volume for '
                               'volume %(volume)s, Exception: %(ex)s.')
                             % {'volume': lun.ds_id, 'ex': six.text_type(e)}))
        lun.replication_status = 'disabled'
        lun.replication_driver_data = {}
        return lun

    @proxy.logger
    def _delete_replica(self, lun):
        if not lun.replication_driver_data:
            LOG.error("No replica ID for lun %s, maybe there is something "
                      "wrong when creating the replica for lun.", lun.ds_id)
            return None

        for backend_id, backend in lun.replication_driver_data.items():
            if not self._mm_manager.is_target_alive():
                return None

            if not self._target_helper.lun_exists(backend['vol_hex_id']):
                LOG.debug("Replica %s not found.", backend['vol_hex_id'])
                continue

            LOG.debug("Deleting replica %s.", backend['vol_hex_id'])
            self._target_helper.delete_lun_by_id(backend['vol_hex_id'])

    def extend_replica(self, lun, param):
        for backend_id, backend in lun.replication_driver_data.items():
            self._target_helper.change_lun(backend['vol_hex_id'], param)

    def delete_pprc_pairs(self, lun):
        self._mm_manager.delete_pprc_pairs(lun)

    def create_pprc_pairs(self, lun):
        self._mm_manager.create_pprc_pairs(lun)

    def start_host_pprc_failover(self, luns, backend_id):
        self._mm_manager.do_pprc_failover(luns)
        self.switch_source_and_target_client()
        self._switch_source_and_target_volume(luns, backend_id)

    def start_group_pprc_failover(self, luns, backend_id):
        self._mm_manager.do_pprc_failover(luns, True)
        self._switch_source_and_target_volume(luns, backend_id)

    def _get_sample_luns(self, luns):
        # choose sample lun according to position.
        sample_luns = []
        positions = []
        for lun in luns:
            position = (lun.pool_lss_pair['source'][1],
                        lun.pool_lss_pair['target'][1])
            if position not in positions:
                sample_luns.append(lun)
                positions.append(position)
        return sample_luns

    @proxy.logger
    def start_host_pprc_failback(self, luns, backend_id):
        # check whether primary client is alive or not.
        if not self._mm_manager.is_target_alive():
            try:
                self._target_helper.update_client()
            except restclient.APIException:
                msg = _("Can not connect to the primary backend, "
                        "please make sure it is back.")
                LOG.error(msg)
                raise exception.UnableToFailOver(reason=msg)
        LOG.debug("Failback host starts, backend id is %s.", backend_id)
        sample_luns = self._get_sample_luns(luns)
        for lun in sample_luns:
            self._mm_manager.create_pprc_path(lun)
        self._mm_manager.do_pprc_failback(luns)
        # revert the relationship of source volume and target volume
        self.start_host_pprc_failover(luns, backend_id)
        self._mm_manager.do_pprc_failback(luns)
        LOG.debug("Failback host ends, backend id is %s.", backend_id)

    @proxy.logger
    def start_group_pprc_failback(self, luns, backend_id):
        LOG.debug("Failback group starts, backend id is %s.", backend_id)
        sample_luns = self._get_sample_luns(luns)
        for lun in sample_luns:
            self._mm_manager.create_pprc_path(lun, True)
        self._mm_manager.do_pprc_failback(luns, True)
        self.start_group_pprc_failover(luns, backend_id)
        self._mm_manager.do_pprc_failback(luns, True)
        LOG.debug("Failback group ends, backend id is %s.", backend_id)

    def _get_expected_luns(self, luns, state, ignored_state=None):
        lun_ids = set(lun.ds_id for lun in luns)
        min_lun_id = min(lun_ids)
        max_lun_id = max(lun_ids)
        if not luns[0].failed_over:
            pairs = self._source_helper.get_pprc_pairs(min_lun_id, max_lun_id)
        else:
            pairs = self._target_helper.get_pprc_pairs(min_lun_id, max_lun_id)
        pairs = {pair['source_volume']['name']: pair for pair in pairs}
        expected_luns = []
        for lun in luns:
            pair = pairs.get(lun.ds_id)
            if pair:
                if ignored_state and pair['state'] == ignored_state:
                    continue
                elif pair['state'] != state:
                    raise exception.VolumeDriverException(
                        message=(_("Source volume %(id)s has wrong pprc pair "
                                   "state %(invalid_state)s, expected one is "
                                   "%(valid_state)s")
                                 % {'id': pair['source_volume']['name'],
                                    'invalid_state': pair['state'],
                                    'valid_state': state}))
            else:
                raise exception.VolumeDriverException(
                    message=_("There is no PPRC pair for source volume "
                              "%s.") % lun.ds_id)
            expected_luns.append(lun)
        return expected_luns

    @proxy.logger
    def enable_replication(self, luns):
        # after group is failed over, user can not enable replication.
        if not luns:
            return None
        luns = self._get_expected_luns(luns, 'suspended', 'full_duplex')
        pprc_pair_ids = self._mm_manager.get_pprc_pair_ids(luns)
        LOG.debug("enable_replication: pprc_pair_ids is %s", pprc_pair_ids)
        if pprc_pair_ids:
            self._source_helper.resume_pprc_pairs(pprc_pair_ids)

    @proxy.logger
    def disable_replication(self, luns):
        # after group is failed over, user can not disable replication.
        if not luns:
            return None
        luns = self._get_expected_luns(luns, 'full_duplex', 'suspended')
        pprc_pair_ids = self._mm_manager.get_pprc_pair_ids(luns)
        LOG.debug("disable_replication: pprc_pair_ids is %s", pprc_pair_ids)
        if pprc_pair_ids:
            self._source_helper.pause_pprc_pairs(pprc_pair_ids)



#    (c) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""HPE LeftHand SAN ISCSI REST Proxy.

Volume driver for HPE LeftHand Storage array.
This driver requires 11.5 or greater firmware on the LeftHand array, using
the 2.0 or greater version of the hpelefthandclient.

You will need to install the python hpelefthandclient module.
sudo pip install python-lefthandclient

Set the following in the cinder.conf file to enable the
LeftHand iSCSI REST Driver along with the required flags:

volume_driver=cinder.volume.drivers.hpe.hpe_lefthand_iscsi.
    HPELeftHandISCSIDriver

It also requires the setting of hpelefthand_api_url, hpelefthand_username,
hpelefthand_password for credentials to talk to the REST service on the
LeftHand array.

"""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils as json
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_utils import units

from cinder import context
from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.objects import fields
from cinder import utils as cinder_utils
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume.drivers.san import san
from cinder.volume import volume_types
from cinder.volume import volume_utils

import math
import re
import six

LOG = logging.getLogger(__name__)

hpelefthandclient = importutils.try_import("hpelefthandclient")
if hpelefthandclient:
    from hpelefthandclient import client as hpe_lh_client
    from hpelefthandclient import exceptions as hpeexceptions

hpelefthand_opts = [
    cfg.URIOpt('hpelefthand_api_url',
               default=None,
               help="HPE LeftHand WSAPI Server Url like "
                    "https://<LeftHand ip>:8081/lhos"),
    cfg.StrOpt('hpelefthand_username',
               default=None,
               help="HPE LeftHand Super user username"),
    cfg.StrOpt('hpelefthand_password',
               default=None,
               help="HPE LeftHand Super user password",
               secret=True),
    cfg.StrOpt('hpelefthand_clustername',
               default=None,
               help="HPE LeftHand cluster name"),
    cfg.BoolOpt('hpelefthand_iscsi_chap_enabled',
                default=False,
                help='Configure CHAP authentication for iSCSI connections '
                '(Default: Disabled)'),
    cfg.BoolOpt('hpelefthand_debug',
                default=False,
                help="Enable HTTP debugging to LeftHand"),
    cfg.PortOpt('hpelefthand_ssh_port',
                default=16022,
                help="Port number of SSH service."),

]

CONF = cfg.CONF
CONF.register_opts(hpelefthand_opts, group=configuration.SHARED_CONF_GROUP)

MIN_API_VERSION = "1.1"
MIN_CLIENT_VERSION = '2.1.0'

# map the extra spec key to the REST client option key
extra_specs_key_map = {
    'hpelh:provisioning': 'isThinProvisioned',
    'hpelh:ao': 'isAdaptiveOptimizationEnabled',
    'hpelh:data_pl': 'dataProtectionLevel',
    'hplh:provisioning': 'isThinProvisioned',
    'hplh:ao': 'isAdaptiveOptimizationEnabled',
    'hplh:data_pl': 'dataProtectionLevel',
}

# map the extra spec value to the REST client option value
extra_specs_value_map = {
    'isThinProvisioned': {'thin': True, 'full': False},
    'isAdaptiveOptimizationEnabled': {'true': True, 'false': False},
    'dataProtectionLevel': {
        'r-0': 0, 'r-5': 1, 'r-10-2': 2, 'r-10-3': 3, 'r-10-4': 4, 'r-6': 5}
}

extra_specs_default_key_value_map = {
    'hpelh:provisioning': 'thin',
    'hpelh:ao': 'true',
    'hpelh:data_pl': 'r-0'
}


@interface.volumedriver
class HPELeftHandISCSIDriver(driver.ISCSIDriver):
    """Executes REST commands relating to HPE/LeftHand SAN ISCSI volumes.

    Version history:

    .. code-block:: none

        1.0.0 - Initial REST iSCSI proxy
        1.0.1 - Added support for retype
        1.0.2 - Added support for volume migrate
        1.0.3 - Fixed bug #1285829, HP LeftHand backend assisted migration
                should check for snapshots
        1.0.4 - Fixed bug #1285925, LeftHand AO volume create performance
                improvement
        1.0.5 - Fixed bug #1311350, Live-migration of an instance when
                attached to a volume was causing an error.
        1.0.6 - Removing locks bug #1395953
        1.0.7 - Fixed bug #1353137, Server was not removed from the HP
                Lefthand backend after the last volume was detached.
        1.0.8 - Fixed bug #1418201, A cloned volume fails to attach.
        1.0.9 - Adding support for manage/unmanage.
        1.0.10 - Add stats for goodness_function and filter_function
        1.0.11 - Add over subscription support
        1.0.12 - Adds consistency group support
        1.0.13 - Added update_migrated_volume #1493546
        1.0.14 - Removed the old CLIQ based driver
        2.0.0 - Rebranded HP to HPE
        2.0.1 - Remove db access for consistency groups
        2.0.2 - Adds v2 managed replication support
        2.0.3 - Adds v2 unmanaged replication support
        2.0.4 - Add manage/unmanage snapshot support
        2.0.5 - Changed minimum client version to be 2.1.0
        2.0.6 - Update replication to version 2.1
        2.0.7 - Fixed bug #1554746, Create clone volume with new size.
        2.0.8 - Add defaults for creating a replication client, bug #1556331
        2.0.9 - Fix terminate connection on failover
        2.0.10 - Add entry point tracing
        2.0.11 - Fix extend volume if larger than snapshot bug #1560654
        2.0.12 - add CG capability to generic volume groups.
        2.0.13 - Fix cloning operation related to provisioning, bug #1688243
        2.0.14 - Fixed bug #1710072, Volume doesn't show expected parameters
                 after Retype
        2.0.15 - Fixed bug #1710098, Managed volume, does not pick up the extra
                 specs/capabilities of the selected volume type.
        2.0.16 - Handled concurrent attachment requests. bug #1779654
    """

    VERSION = "2.0.16"

    CI_WIKI_NAME = "HPE_Storage_CI"

    # TODO(walter-boring) Remove this driver in the 'U' release as it is no
    # longer maintained.
    SUPPORTED = False

    device_stats = {}

    # v2 replication constants
    EXTRA_SPEC_REP_SYNC_PERIOD = "replication:sync_period"
    EXTRA_SPEC_REP_RETENTION_COUNT = "replication:retention_count"
    EXTRA_SPEC_REP_REMOTE_RETENTION_COUNT = (
        "replication:remote_retention_count")
    MIN_REP_SYNC_PERIOD = 1800
    DEFAULT_RETENTION_COUNT = 5
    MAX_RETENTION_COUNT = 50
    DEFAULT_REMOTE_RETENTION_COUNT = 5
    MAX_REMOTE_RETENTION_COUNT = 50
    REP_SNAPSHOT_SUFFIX = "_SS"
    REP_SCHEDULE_SUFFIX = "_SCHED"
    FAILBACK_VALUE = 'default'

    def __init__(self, *args, **kwargs):
        super(HPELeftHandISCSIDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(hpelefthand_opts)
        self.configuration.append_config_values(san.san_opts)
        if not self.configuration.hpelefthand_api_url:
            raise exception.NotFound(_("HPELeftHand url not found"))

        # blank is the only invalid character for cluster names
        # so we need to use it as a separator
        self.DRIVER_LOCATION = self.__class__.__name__ + ' %(cluster)s %(vip)s'
        self._client_conf = {}
        self._replication_targets = []
        self._replication_enabled = False
        self._active_backend_id = kwargs.get('active_backend_id', None)

    @staticmethod
    def get_driver_options():
        return hpelefthand_opts

    def _login(self, timeout=None):
        conf = self._get_lefthand_config()
        if conf:
            self._client_conf['hpelefthand_username'] = (
                conf['hpelefthand_username'])
            self._client_conf['hpelefthand_password'] = (
                conf['hpelefthand_password'])
            self._client_conf['hpelefthand_clustername'] = (
                conf['hpelefthand_clustername'])
            self._client_conf['hpelefthand_api_url'] = (
                conf['hpelefthand_api_url'])
            self._client_conf['hpelefthand_ssh_port'] = (
                conf['hpelefthand_ssh_port'])
            self._client_conf['hpelefthand_iscsi_chap_enabled'] = (
                conf['hpelefthand_iscsi_chap_enabled'])
            self._client_conf['ssh_conn_timeout'] = conf['ssh_conn_timeout']
            self._client_conf['san_private_key'] = conf['san_private_key']
        else:
            self._client_conf['hpelefthand_username'] = (
                self.configuration.hpelefthand_username)
            self._client_conf['hpelefthand_password'] = (
                self.configuration.hpelefthand_password)
            self._client_conf['hpelefthand_clustername'] = (
                self.configuration.hpelefthand_clustername)
            self._client_conf['hpelefthand_api_url'] = (
                self.configuration.hpelefthand_api_url)
            self._client_conf['hpelefthand_ssh_port'] = (
                self.configuration.hpelefthand_ssh_port)
            self._client_conf['hpelefthand_iscsi_chap_enabled'] = (
                self.configuration.hpelefthand_iscsi_chap_enabled)
            self._client_conf['ssh_conn_timeout'] = (
                self.configuration.ssh_conn_timeout)
            self._client_conf['san_private_key'] = (
                self.configuration.san_private_key)

        client = self._create_client(timeout=timeout)
        try:
            if self.configuration.hpelefthand_debug:
                client.debug_rest(True)

            client.login(
                self._client_conf['hpelefthand_username'],
                self._client_conf['hpelefthand_password'])

            cluster_info = client.getClusterByName(
                self._client_conf['hpelefthand_clustername'])
            self.cluster_id = cluster_info['id']
            virtual_ips = cluster_info['virtualIPAddresses']
            self.cluster_vip = virtual_ips[0]['ipV4Address']

            # Extract IP address from API URL
            ssh_ip = self._extract_ip_from_url(
                self._client_conf['hpelefthand_api_url'])
            known_hosts_file = CONF.ssh_hosts_key_file
            policy = "AutoAddPolicy"
            if CONF.strict_ssh_host_key_policy:
                policy = "RejectPolicy"
            client.setSSHOptions(
                ssh_ip,
                self._client_conf['hpelefthand_username'],
                self._client_conf['hpelefthand_password'],
                port=self._client_conf['hpelefthand_ssh_port'],
                conn_timeout=self._client_conf['ssh_conn_timeout'],
                privatekey=self._client_conf['san_private_key'],
                missing_key_policy=policy,
                known_hosts_file=known_hosts_file)

            return client
        except hpeexceptions.HTTPNotFound:
            raise exception.DriverNotInitialized(
                _('LeftHand cluster not found'))
        except Exception as ex:
            raise exception.DriverNotInitialized(ex)

    def _logout(self, client):
        if client is not None:
            client.logout()

    def _create_client(self, timeout=None):
        # Timeout is only supported in version 2.0.1 and greater of the
        # python-lefthandclient.
        hpelefthand_api_url = self._client_conf['hpelefthand_api_url']
        client = hpe_lh_client.HPELeftHandClient(
            hpelefthand_api_url, timeout=timeout)
        return client

    def _create_replication_client(self, remote_array):
        cl = hpe_lh_client.HPELeftHandClient(
            remote_array['hpelefthand_api_url'])
        try:
            cl.login(
                remote_array['hpelefthand_username'],
                remote_array['hpelefthand_password'])

            ssh_conn_timeout = remote_array.get('ssh_conn_timeout', 30)
            san_private_key = remote_array.get('san_private_key', '')

            # Extract IP address from API URL
            ssh_ip = self._extract_ip_from_url(
                remote_array['hpelefthand_api_url'])
            known_hosts_file = CONF.ssh_hosts_key_file
            policy = "AutoAddPolicy"
            if CONF.strict_ssh_host_key_policy:
                policy = "RejectPolicy"
            cl.setSSHOptions(
                ssh_ip,
                remote_array['hpelefthand_username'],
                remote_array['hpelefthand_password'],
                port=remote_array['hpelefthand_ssh_port'],
                conn_timeout=ssh_conn_timeout,
                privatekey=san_private_key,
                missing_key_policy=policy,
                known_hosts_file=known_hosts_file)

            return cl
        except hpeexceptions.HTTPNotFound:
            raise exception.DriverNotInitialized(
                _('LeftHand cluster not found'))
        except Exception as ex:
            raise exception.DriverNotInitialized(ex)

    def _destroy_replication_client(self, client):
        if client is not None:
            client.logout()

    def _extract_ip_from_url(self, url):
        result = re.search("://(.*):", url)
        ip = result.group(1)
        return ip

    def do_setup(self, context):
        """Set up LeftHand client."""
        if not hpelefthandclient:
            # Checks if client was successfully imported
            ex_msg = _("HPELeftHand client is not installed. Please"
                       " install using 'pip install "
                       "python-lefthandclient'.")
            LOG.error(ex_msg)
            raise exception.VolumeDriverException(ex_msg)

        if hpelefthandclient.version < MIN_CLIENT_VERSION:
            ex_msg = (_("Invalid hpelefthandclient version found ("
                        "%(found)s). Version %(minimum)s or greater "
                        "required. Run 'pip install --upgrade "
                        "python-lefthandclient' to upgrade the "
                        "hpelefthandclient.")
                      % {'found': hpelefthandclient.version,
                         'minimum': MIN_CLIENT_VERSION})
            LOG.error(ex_msg)
            raise exception.InvalidInput(reason=ex_msg)

        self._do_replication_setup()

    def check_for_setup_error(self):
        """Checks for incorrect LeftHand API being used on backend."""
        client = self._login()
        try:
            self.api_version = client.getApiVersion()

            LOG.info("HPELeftHand API version %s", self.api_version)

            if self.api_version < MIN_API_VERSION:
                LOG.warning("HPELeftHand API is version %(current)s. "
                            "A minimum version of %(min)s is needed for "
                            "manage/unmanage support.",
                            {'current': self.api_version,
                             'min': MIN_API_VERSION})
        finally:
            self._logout(client)

    def check_replication_flags(self, options, required_flags):
        for flag in required_flags:
            if not options.get(flag, None):
                msg = _('%s is not set and is required for the replication '
                        'device to be valid.') % flag
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)

    def get_version_string(self):
        return (_('REST %(proxy_ver)s hpelefthandclient %(rest_ver)s') % {
            'proxy_ver': self.VERSION,
            'rest_ver': hpelefthandclient.get_version_string()})

    @cinder_utils.trace
    def create_volume(self, volume):
        """Creates a volume."""
        client = self._login()
        try:
            # get the extra specs of interest from this volume's volume type
            volume_extra_specs = self._get_volume_extra_specs(volume)
            extra_specs = self._get_lh_extra_specs(
                volume_extra_specs,
                extra_specs_key_map.keys())

            # map the extra specs key/value pairs to key/value pairs
            # used as optional configuration values by the LeftHand backend
            optional = self._map_extra_specs(extra_specs)

            # if provisioning is not set, default to thin
            if 'isThinProvisioned' not in optional:
                optional['isThinProvisioned'] = True

            # AdaptiveOptimization defaults to 'true' if you don't specify the
            # value on a create, and that is the most efficient way to create
            # a volume. If you pass in 'false' or 'true' for AO, it will result
            # in an update operation following the create operation to set this
            # value, so it is best to not specify the value and let it default
            # to 'true'.
            if optional.get('isAdaptiveOptimizationEnabled'):
                del optional['isAdaptiveOptimizationEnabled']

            clusterName = self._client_conf['hpelefthand_clustername']
            optional['clusterName'] = clusterName

            volume_info = client.createVolume(
                volume['name'], self.cluster_id,
                volume['size'] * units.Gi,
                optional)

            model_update = self._update_provider(volume_info)

            # v2 replication check
            if self._volume_of_replicated_type(volume) and (
               self._do_volume_replication_setup(volume, client, optional)):
                model_update['replication_status'] = 'enabled'
                model_update['replication_driver_data'] = (json.dumps(
                    {'location': self._client_conf['hpelefthand_api_url']}))

            return model_update
        except Exception as ex:
            raise exception.VolumeBackendAPIException(data=ex)
        finally:
            self._logout(client)

    @cinder_utils.trace
    def delete_volume(self, volume):
        """Deletes a volume."""
        client = self._login()
        # v2 replication check
        # If the volume type is replication enabled, we want to call our own
        # method of deconstructing the volume and its dependencies
        if self._volume_of_replicated_type(volume):
            self._do_volume_replication_destroy(volume, client)
            return

        try:
            volume_info = client.getVolumeByName(volume['name'])
            client.deleteVolume(volume_info['id'])
        except hpeexceptions.HTTPNotFound:
            LOG.error("Volume did not exist. It will not be deleted")
        except Exception as ex:
            raise exception.VolumeBackendAPIException(ex)
        finally:
            self._logout(client)

    @cinder_utils.trace
    def extend_volume(self, volume, new_size):
        """Extend the size of an existing volume."""
        client = self._login()
        try:
            volume_info = client.getVolumeByName(volume['name'])

            # convert GB to bytes
            options = {'size': int(new_size) * units.Gi}
            client.modifyVolume(volume_info['id'], options)
        except Exception as ex:
            raise exception.VolumeBackendAPIException(ex)
        finally:
            self._logout(client)

    @cinder_utils.trace
    def create_group(self, context, group):
        """Creates a group."""
        LOG.debug("Creating group.")
        if not volume_utils.is_group_a_cg_snapshot_type(group):
            raise NotImplementedError()
        for vol_type_id in group.volume_type_ids:
            replication_type = self._volume_of_replicated_type(
                None, vol_type_id)
            if replication_type:
                # An unsupported configuration
                LOG.error('Unable to create group: create group with '
                          'replication volume type is not supported.')
                model_update = {'status': fields.GroupStatus.ERROR}
                return model_update

        return {'status': fields.GroupStatus.AVAILABLE}

    @cinder_utils.trace
    def create_group_from_src(self, context, group, volumes,
                              group_snapshot=None, snapshots=None,
                              source_group=None, source_vols=None):
        """Creates a group from a source"""
        msg = _("Creating a group from a source is not "
                "supported when consistent_group_snapshot_enabled to true.")
        if not volume_utils.is_group_a_cg_snapshot_type(group):
            raise NotImplementedError()
        else:
            raise exception.VolumeBackendAPIException(data=msg)

    @cinder_utils.trace
    def delete_group(self, context, group, volumes):
        """Deletes a group."""
        if not volume_utils.is_group_a_cg_snapshot_type(group):
            raise NotImplementedError()

        volume_model_updates = []
        for volume in volumes:
            volume_update = {'id': volume.id}
            try:
                self.delete_volume(volume)
                volume_update['status'] = 'deleted'
            except Exception as ex:
                LOG.error("There was an error deleting volume %(id)s: "
                          "%(error)s.",
                          {'id': volume.id,
                           'error': ex})
                volume_update['status'] = 'error'
            volume_model_updates.append(volume_update)

        model_update = {'status': group.status}

        return model_update, volume_model_updates

    @cinder_utils.trace
    def update_group(self, context, group, add_volumes=None,
                     remove_volumes=None):
        """Updates a group.

        Because the backend has no concept of volume grouping, cinder will
        maintain all volume/group relationships. Because of this
        functionality, there is no need to make any client calls; instead
        simply returning out of this function allows cinder to properly
        add/remove volumes from the group.
        """
        LOG.debug("Updating group.")
        if not volume_utils.is_group_a_cg_snapshot_type(group):
            raise NotImplementedError()

        return None, None, None

    @cinder_utils.trace
    def create_group_snapshot(self, context, group_snapshot, snapshots):
        """Creates a group snapshot."""
        if not volume_utils.is_group_a_cg_snapshot_type(group_snapshot):
            raise NotImplementedError()
        client = self._login()
        try:
            snap_set = []
            snapshot_base_name = "snapshot-" + group_snapshot.id
            snapshot_model_updates = []
            for i, snapshot in enumerate(snapshots):
                volume = snapshot.volume
                volume_name = volume['name']
                try:
                    volume_info = client.getVolumeByName(volume_name)
                except Exception as ex:
                    error = six.text_type(ex)
                    LOG.error("Could not find volume with name %(name)s. "
                              "Error: %(error)s",
                              {'name': volume_name,
                               'error': error})
                    raise exception.VolumeBackendAPIException(data=error)

                volume_id = volume_info['id']
                snapshot_name = snapshot_base_name + "-" + six.text_type(i)
                snap_set_member = {'volumeName': volume_name,
                                   'volumeId': volume_id,
                                   'snapshotName': snapshot_name}
                snap_set.append(snap_set_member)
                snapshot_update = {'id': snapshot['id'],
                                   'status': fields.SnapshotStatus.AVAILABLE}
                snapshot_model_updates.append(snapshot_update)

            source_volume_id = snap_set[0]['volumeId']
            optional = {'inheritAccess': True}
            description = group_snapshot.description
            if description:
                optional['description'] = description

            try:
                client.createSnapshotSet(source_volume_id, snap_set, optional)
            except Exception as ex:
                error = six.text_type(ex)
                LOG.error("Could not create snapshot set. Error: '%s'",
                          error)
                raise exception.VolumeBackendAPIException(
                    data=error)

        except Exception as ex:
            raise exception.VolumeBackendAPIException(data=six.text_type(ex))
        finally:
            self._logout(client)

        model_update = {'status': 'available'}

        return model_update, snapshot_model_updates

    @cinder_utils.trace
    def delete_group_snapshot(self, context, group_snapshot, snapshots):
        """Deletes a group snapshot."""
        if not volume_utils.is_group_a_cg_snapshot_type(group_snapshot):
            raise NotImplementedError()
        client = self._login()
        snap_name_base = "snapshot-" + group_snapshot.id

        snapshot_model_updates = []
        for i, snapshot in enumerate(snapshots):
            snapshot_update = {'id': snapshot['id']}
            try:
                snap_name = snap_name_base + "-" + six.text_type(i)
                snap_info = client.getSnapshotByName(snap_name)
                client.deleteSnapshot(snap_info['id'])
                snapshot_update['status'] = fields.SnapshotStatus.DELETED
            except hpeexceptions.HTTPServerError as ex:
                in_use_msg = ('cannot be deleted because it is a clone '
                              'point')
                if in_use_msg in ex.get_description():
                    LOG.error("The snapshot cannot be deleted because "
                              "it is a clone point.")
                snapshot_update['status'] = fields.SnapshotStatus.ERROR
            except Exception as ex:
                LOG.error("There was an error deleting snapshot %(id)s: "
                          "%(error)s.",
                          {'id': snapshot['id'],
                           'error': six.text_type(ex)})
                snapshot_update['status'] = fields.SnapshotStatus.ERROR
            snapshot_model_updates.append(snapshot_update)

        self._logout(client)

        model_update = {'status': group_snapshot.status}

        return model_update, snapshot_model_updates

    @cinder_utils.trace
    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        client = self._login()
        try:
            volume_info = client.getVolumeByName(snapshot['volume_name'])

            option = {'inheritAccess': True}
            client.createSnapshot(snapshot['name'],
                                  volume_info['id'],
                                  option)
        except Exception as ex:
            raise exception.VolumeBackendAPIException(ex)
        finally:
            self._logout(client)

    @cinder_utils.trace
    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        client = self._login()
        try:
            snap_info = client.getSnapshotByName(snapshot['name'])
            client.deleteSnapshot(snap_info['id'])
        except hpeexceptions.HTTPNotFound:
            LOG.error("Snapshot did not exist. It will not be deleted")
        except hpeexceptions.HTTPServerError as ex:
            in_use_msg = 'cannot be deleted because it is a clone point'
            if in_use_msg in ex.get_description():
                raise exception.SnapshotIsBusy(snapshot_name=snapshot['name'])

            raise exception.VolumeBackendAPIException(ex)

        except Exception as ex:
            raise exception.VolumeBackendAPIException(ex)
        finally:
            self._logout(client)

    @cinder_utils.trace
    def get_volume_stats(self, refresh=False):
        """Gets volume stats."""
        client = self._login()
        try:
            if refresh:
                self._update_backend_status(client)

            return self.device_stats
        finally:
            self._logout(client)

    def _update_backend_status(self, client):
        data = {}
        backend_name = self.configuration.safe_get('volume_backend_name')
        data['driver_version'] = self.VERSION
        data['volume_backend_name'] = backend_name or self.__class__.__name__
        data['reserved_percentage'] = (
            self.configuration.safe_get('reserved_percentage'))
        data['storage_protocol'] = 'iSCSI'
        data['vendor_name'] = 'Hewlett Packard Enterprise'
        data['location_info'] = (self.DRIVER_LOCATION % {
            'cluster': self._client_conf['hpelefthand_clustername'],
            'vip': self.cluster_vip})
        data['thin_provisioning_support'] = True
        data['thick_provisioning_support'] = True
        data['max_over_subscription_ratio'] = (
            self.configuration.safe_get('max_over_subscription_ratio'))

        cluster_info = client.getCluster(self.cluster_id)

        total_capacity = cluster_info['spaceTotal']
        free_capacity = cluster_info['spaceAvailable']

        # convert to GB
        data['total_capacity_gb'] = int(total_capacity) / units.Gi
        data['free_capacity_gb'] = int(free_capacity) / units.Gi

        # Collect some stats
        capacity_utilization = (
            (float(total_capacity - free_capacity) /
             float(total_capacity)) * 100)
        # Don't have a better way to get the total number volumes
        # so try to limit the size of data for now. Once new lefthand API is
        # available, replace this call.
        total_volumes = 0
        provisioned_size = 0
        volumes = client.getVolumes(
            cluster=self._client_conf['hpelefthand_clustername'],
            fields=['members[id]', 'members[clusterName]', 'members[size]'])
        if volumes:
            total_volumes = volumes['total']
            provisioned_size = sum(
                members['size'] for members in volumes['members'])
        data['provisioned_capacity_gb'] = int(provisioned_size) / units.Gi
        data['capacity_utilization'] = capacity_utilization
        data['total_volumes'] = total_volumes
        data['filter_function'] = self.get_filter_function()
        data['goodness_function'] = self.get_goodness_function()
        data['consistent_group_snapshot_enabled'] = True
        data['replication_enabled'] = self._replication_enabled
        data['replication_type'] = ['periodic']
        data['replication_count'] = len(self._replication_targets)
        data['replication_targets'] = self._get_replication_targets()

        self.device_stats = data

    @cinder_utils.trace
    def initialize_connection(self, volume, connector):
        """Assigns the volume to a server.

        Assign any created volume to a compute node/host so that it can be
        used from that host. HPE VSA requires a volume to be assigned
        to a server.
        """
        client = self._login()
        try:
            server_info = self._create_server(connector, client)
            volume_info = client.getVolumeByName(volume['name'])

            access_already_enabled = False
            if volume_info['iscsiSessions'] is not None:
                # Extract the server id for each session to check if the
                # new server already has access permissions enabled.
                for session in volume_info['iscsiSessions']:
                    server_id = int(session['server']['uri'].split('/')[3])
                    if server_id == server_info['id']:
                        access_already_enabled = True
                        break

            if not access_already_enabled:
                client.addServerAccess(
                    volume_info['id'],
                    server_info['id'])

            iscsi_properties = self._get_iscsi_properties(volume)

            if ('chapAuthenticationRequired' in server_info and
                    server_info['chapAuthenticationRequired']):
                iscsi_properties['auth_method'] = 'CHAP'
                iscsi_properties['auth_username'] = connector['initiator']
                iscsi_properties['auth_password'] = (
                    server_info['chapTargetSecret'])

            return {'driver_volume_type': 'iscsi', 'data': iscsi_properties}
        except Exception as ex:
            raise exception.VolumeBackendAPIException(ex)
        finally:
            self._logout(client)

    @cinder_utils.trace
    def terminate_connection(self, volume, connector, **kwargs):
        """Unassign the volume from the host."""
        client = self._login()
        try:
            volume_info = client.getVolumeByName(volume['name'])
            server_info = client.getServerByName(connector['host'])
            volume_list = client.findServerVolumes(server_info['name'])

            removeServer = True
            for entry in volume_list:
                if entry['id'] != volume_info['id']:
                    removeServer = False
                    break

            client.removeServerAccess(
                volume_info['id'],
                server_info['id'])

            if removeServer:
                client.deleteServer(server_info['id'])
        except hpeexceptions.HTTPNotFound as ex:
            # If a host is failed-over, we want to allow the detach to
            # to 'succeed' when it cannot find the host. We can simply
            # return out of the terminate connection in order for things
            # to be updated correctly.
            if self._active_backend_id:
                LOG.warning("Because the host is currently in a "
                            "failed-over state, the volume will not "
                            "be properly detached from the primary "
                            "array. The detach will be considered a "
                            "success as far as Cinder is concerned. "
                            "The volume can now be attached to the "
                            "secondary target.")
                return
            else:
                raise exception.VolumeBackendAPIException(ex)
        except Exception as ex:
            raise exception.VolumeBackendAPIException(ex)
        finally:
            self._logout(client)

    @cinder_utils.trace
    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""
        client = self._login()
        try:
            snap_info = client.getSnapshotByName(snapshot['name'])
            volume_info = client.cloneSnapshot(
                volume['name'],
                snap_info['id'])

            # Extend volume
            if volume['size'] > snapshot['volume_size']:
                LOG.debug("Resize the new volume to %s.", volume['size'])
                self.extend_volume(volume, volume['size'])

            model_update = self._update_provider(volume_info)

            # v2 replication check
            if self._volume_of_replicated_type(volume) and (
               self._do_volume_replication_setup(volume, client)):
                model_update['replication_status'] = 'enabled'
                model_update['replication_driver_data'] = (json.dumps(
                    {'location': self._client_conf['hpelefthand_api_url']}))

            return model_update
        except Exception as ex:
            raise exception.VolumeBackendAPIException(ex)
        finally:
            self._logout(client)

    @cinder_utils.trace
    def create_cloned_volume(self, volume, src_vref):
        client = self._login()
        try:
            volume_info = client.getVolumeByName(src_vref['name'])
            clone_info = client.cloneVolume(volume['name'], volume_info['id'])

            # Extend volume
            if volume['size'] > src_vref['size']:
                LOG.debug("Resize the new volume to %s.", volume['size'])
                self.extend_volume(volume, volume['size'])
            # TODO(kushal) : we will use volume.volume_types when we re-write
            # the design for unit tests to use objects instead of dicts.
            # Get the extra specs of interest from this volume's volume type
            volume_extra_specs = self._get_volume_extra_specs(src_vref)
            extra_specs = self._get_lh_extra_specs(
                volume_extra_specs,
                extra_specs_key_map.keys())

            # Check provisioning type of source volume. If it's full then need
            # to change provisioning of clone volume to full as lefthand
            # creates clone volume only with thin provisioning type.
            if extra_specs.get('hpelh:provisioning') == 'full':
                options = {'isThinProvisioned': False}
                clone_volume_info = client.getVolumeByName(volume['name'])
                client.modifyVolume(clone_volume_info['id'], options)

            model_update = self._update_provider(clone_info)

            # v2 replication check
            if self._volume_of_replicated_type(volume) and (
               self._do_volume_replication_setup(volume, client)):
                model_update['replication_status'] = 'enabled'
                model_update['replication_driver_data'] = (json.dumps(
                    {'location': self._client_conf['hpelefthand_api_url']}))

            return model_update
        except Exception as ex:
            raise exception.VolumeBackendAPIException(ex)
        finally:
            self._logout(client)

    def _get_volume_extra_specs(self, volume):
        """Get extra specs from a volume."""
        extra_specs = {}
        type_id = volume.get('volume_type_id', None)
        if type_id is not None:
            ctxt = context.get_admin_context()
            volume_type = volume_types.get_volume_type(ctxt, type_id)
            extra_specs = volume_type.get('extra_specs')
        return extra_specs

    def _get_lh_extra_specs(self, extra_specs, valid_keys):
        """Get LeftHand extra_specs (valid_keys only)."""
        extra_specs_of_interest = {}
        for key, value in extra_specs.items():
            if key in valid_keys:
                prefix = key.split(":")
                if prefix[0] == "hplh":
                    LOG.warning("The 'hplh' prefix is deprecated. Use "
                                "'hpelh' instead.")
                extra_specs_of_interest[key] = value
        return extra_specs_of_interest

    def _map_extra_specs(self, extra_specs):
        """Map the extra spec key/values to LeftHand key/values."""
        client_options = {}
        for key, value in extra_specs.items():
            # map extra spec key to lh client option key
            client_key = extra_specs_key_map[key]
            # map extra spect value to lh client option value
            try:
                value_map = extra_specs_value_map[client_key]
                # an invalid value will throw KeyError
                client_value = value_map[value]
                client_options[client_key] = client_value
            except KeyError:
                LOG.error("'%(value)s' is an invalid value "
                          "for extra spec '%(key)s'",
                          {'value': value, 'key': key})
        return client_options

    def _update_provider(self, volume_info, cluster_vip=None):
        if not cluster_vip:
            cluster_vip = self.cluster_vip
        # TODO(justinsb): Is this always 1? Does it matter?
        cluster_interface = '1'
        iscsi_portal = cluster_vip + ":3260," + cluster_interface

        return {'provider_location': (
            "%s %s %s" % (iscsi_portal, volume_info['iscsiIqn'], 0))}

    @coordination.synchronized('VSA-{connector[host]}')
    def _create_server(self, connector, client):
        server_info = None
        chap_enabled = self._client_conf.get('hpelefthand_iscsi_chap_enabled')
        try:
            server_info = client.getServerByName(connector['host'])
            chap_secret = server_info['chapTargetSecret']
            if not chap_enabled and chap_secret:
                LOG.warning('CHAP secret exists for host %s but CHAP is '
                            'disabled', connector['host'])
            if chap_enabled and chap_secret is None:
                LOG.warning('CHAP is enabled, but server secret not '
                            'configured on server %s', connector['host'])
            return server_info
        except hpeexceptions.HTTPNotFound:
            # server does not exist, so create one
            pass

        optional = None
        if chap_enabled:
            chap_secret = volume_utils.generate_password()
            optional = {'chapName': connector['initiator'],
                        'chapTargetSecret': chap_secret,
                        'chapAuthenticationRequired': True
                        }

        server_info = client.createServer(connector['host'],
                                          connector['initiator'],
                                          optional)
        return server_info

    def create_export(self, context, volume, connector):
        pass

    def ensure_export(self, context, volume):
        pass

    def remove_export(self, context, volume):
        pass

    @cinder_utils.trace
    def retype(self, ctxt, volume, new_type, diff, host):
        """Convert the volume to be of the new type.

        Returns a boolean indicating whether the retype occurred.

        :param ctxt: Context
        :param volume: A dictionary describing the volume to retype
        :param new_type: A dictionary describing the volume type to convert to
        :param diff: A dictionary with the difference between the two types
        :param host: A dictionary describing the host, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        """
        LOG.debug('enter: retype: id=%(id)s, new_type=%(new_type)s,'
                  'diff=%(diff)s, host=%(host)s', {'id': volume['id'],
                                                   'new_type': new_type,
                                                   'diff': diff,
                                                   'host': host})
        client = self._login()
        try:
            volume_info = client.getVolumeByName(volume['name'])

            # pick out the LH extra specs
            new_extra_specs = dict(new_type).get('extra_specs')

            # in the absence of LH capability in diff,
            # True should be return as retype is not needed
            if not list(filter((lambda key: extra_specs_key_map.get(key)),
                               diff['extra_specs'].keys())):
                return True

            # add capability of LH, which are absent in new type,
            # so default value gets set for those capability
            for key, value in extra_specs_default_key_value_map.items():
                if key not in new_extra_specs.keys():
                    new_extra_specs[key] = value

            lh_extra_specs = self._get_lh_extra_specs(
                new_extra_specs,
                extra_specs_key_map.keys())

            LOG.debug('LH specs=%(specs)s', {'specs': lh_extra_specs})

            # only set the ones that have changed
            changed_extra_specs = {}
            for key, value in lh_extra_specs.items():
                try:
                    (old, new) = diff['extra_specs'][key]
                    if old != new:
                        changed_extra_specs[key] = value
                except KeyError:
                    changed_extra_specs[key] = value

            # map extra specs to LeftHand options
            options = self._map_extra_specs(changed_extra_specs)
            if len(options) > 0:
                client.modifyVolume(volume_info['id'], options)
            return True
        except hpeexceptions.HTTPNotFound:
            raise exception.VolumeNotFound(volume_id=volume['id'])
        except Exception as ex:
            LOG.warning("%s", ex)
        finally:
            self._logout(client)

        return False

    @cinder_utils.trace
    def migrate_volume(self, ctxt, volume, host):
        """Migrate the volume to the specified host.

        Backend assisted volume migration will occur if and only if;

        1. Same LeftHand backend
        2. Volume cannot be attached
        3. Volumes with snapshots cannot be migrated
        4. Source and Destination clusters must be in the same management group

        Volume re-type is not supported.

        Returns a boolean indicating whether the migration occurred, as well as
        model_update.

        :param ctxt: Context
        :param volume: A dictionary describing the volume to migrate
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        """
        false_ret = (False, None)
        if 'location_info' not in host['capabilities']:
            return false_ret

        host_location = host['capabilities']['location_info']
        (driver, cluster, vip) = host_location.split(' ')
        client = self._login()
        LOG.debug('enter: migrate_volume: id=%(id)s, host=%(host)s, '
                  'cluster=%(cluster)s', {
                      'id': volume['id'],
                      'host': host,
                      'cluster': self._client_conf['hpelefthand_clustername']})
        try:
            # get the cluster info, if it exists and compare
            cluster_info = client.getClusterByName(cluster)
            LOG.debug('Cluster info: %s', cluster_info)
            virtual_ips = cluster_info['virtualIPAddresses']

            if driver != self.__class__.__name__:
                LOG.info("Cannot provide backend assisted migration for "
                         "volume: %s because volume is from a different "
                         "backend.", volume['name'])
                return false_ret
            if vip != virtual_ips[0]['ipV4Address']:
                LOG.info("Cannot provide backend assisted migration for "
                         "volume: %s because cluster exists in different "
                         "management group.", volume['name'])
                return false_ret

        except hpeexceptions.HTTPNotFound:
            LOG.info("Cannot provide backend assisted migration for "
                     "volume: %s because cluster exists in different "
                     "management group.", volume['name'])
            return false_ret
        finally:
            self._logout(client)

        client = self._login()
        try:
            volume_info = client.getVolumeByName(volume['name'])
            LOG.debug('Volume info: %s', volume_info)

            # can't migrate if server is attached
            if volume_info['iscsiSessions'] is not None:
                LOG.info("Cannot provide backend assisted migration "
                         "for volume: %s because the volume has been "
                         "exported.", volume['name'])
                return false_ret

            # can't migrate if volume has snapshots
            snap_info = client.getVolume(
                volume_info['id'],
                'fields=snapshots,snapshots[resource[members[name]]]')
            LOG.debug('Snapshot info: %s', snap_info)
            if snap_info['snapshots']['resource'] is not None:
                LOG.info("Cannot provide backend assisted migration "
                         "for volume: %s because the volume has "
                         "snapshots.", volume['name'])
                return false_ret

            options = {'clusterName': cluster}
            client.modifyVolume(volume_info['id'], options)
        except hpeexceptions.HTTPNotFound:
            LOG.info("Cannot provide backend assisted migration for "
                     "volume: %s because volume does not exist in this "
                     "management group.", volume['name'])
            return false_ret
        except hpeexceptions.HTTPServerError as ex:
            LOG.error("Exception: %s", ex)
            return false_ret
        finally:
            self._logout(client)

        return (True, None)

    @cinder_utils.trace
    def update_migrated_volume(self, context, volume, new_volume,
                               original_volume_status):
        """Rename the new (temp) volume to it's original name.


        This method tries to rename the new volume to it's original
        name after the migration has completed.

        """
        LOG.debug("Update volume name for %(id)s.", {'id': new_volume['id']})
        name_id = None
        provider_location = None
        if original_volume_status == 'available':
            # volume isn't attached and can be updated
            original_name = CONF.volume_name_template % volume['id']
            current_name = CONF.volume_name_template % new_volume['id']
            client = self._login()
            try:
                volume_info = client.getVolumeByName(current_name)
                volumeMods = {'name': original_name}
                client.modifyVolume(volume_info['id'], volumeMods)
                LOG.info("Volume name changed from %(tmp)s to %(orig)s.",
                         {'tmp': current_name, 'orig': original_name})
            except Exception as e:
                LOG.error("Changing the volume name from %(tmp)s to "
                          "%(orig)s failed because %(reason)s.",
                          {'tmp': current_name, 'orig': original_name,
                           'reason': e})
                name_id = new_volume['_name_id'] or new_volume['id']
                provider_location = new_volume['provider_location']
            finally:
                self._logout(client)
        else:
            # the backend can't change the name.
            name_id = new_volume['_name_id'] or new_volume['id']
            provider_location = new_volume['provider_location']

        return {'_name_id': name_id, 'provider_location': provider_location}

    @cinder_utils.trace
    def manage_existing(self, volume, existing_ref):
        """Manage an existing LeftHand volume.

        existing_ref is a dictionary of the form:
        {'source-name': <name of the virtual volume>}
        """
        # Check API Version
        self._check_api_version()

        target_vol_name = self._get_existing_volume_ref_name(existing_ref)

        # Check for the existence of the virtual volume.
        client = self._login()
        try:
            volume_info = client.getVolumeByName(target_vol_name)
        except hpeexceptions.HTTPNotFound:
            err = (_("Virtual volume '%s' doesn't exist on array.") %
                   target_vol_name)
            LOG.error(err)
            raise exception.InvalidInput(reason=err)
        finally:
            self._logout(client)

        # Generate the new volume information based on the new ID.
        new_vol_name = 'volume-' + volume['id']

        volume_type = None
        if volume['volume_type_id']:
            try:
                volume_type = self._get_volume_type(volume['volume_type_id'])
            except Exception:
                reason = (_("Volume type ID '%s' is invalid.") %
                          volume['volume_type_id'])
                raise exception.ManageExistingVolumeTypeMismatch(reason=reason)

        new_vals = {"name": new_vol_name}

        client = self._login()
        try:
            # Update the existing volume with the new name.
            client.modifyVolume(volume_info['id'], new_vals)
        finally:
            self._logout(client)

        LOG.info("Virtual volume '%(ref)s' renamed to '%(new)s'.",
                 {'ref': existing_ref['source-name'], 'new': new_vol_name})

        display_name = None
        if volume['display_name']:
            display_name = volume['display_name']

        if volume_type:
            LOG.info("Virtual volume %(disp)s '%(new)s' is being retyped.",
                     {'disp': display_name, 'new': new_vol_name})

            # Creates a diff as it needed for retype operation.
            diff = {}
            diff['extra_specs'] = {key: (None, value) for key, value
                                   in volume_type['extra_specs'].items()}
            try:
                self.retype(None,
                            volume,
                            volume_type,
                            diff,
                            volume['host'])
                LOG.info("Virtual volume %(disp)s successfully retyped to "
                         "%(new_type)s.",
                         {'disp': display_name,
                          'new_type': volume_type.get('name')})
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.warning("Failed to manage virtual volume %(disp)s "
                                "due to error during retype.",
                                {'disp': display_name})
                    # Try to undo the rename and clear the new comment.
                    client = self._login()
                    try:
                        client.modifyVolume(
                            volume_info['id'],
                            {'name': target_vol_name})
                    finally:
                        self._logout(client)

        updates = {'display_name': display_name}

        LOG.info("Virtual volume %(disp)s '%(new)s' is now being managed.",
                 {'disp': display_name, 'new': new_vol_name})

        # Return display name to update the name displayed in the GUI and
        # any model updates from retype.
        return updates

    @cinder_utils.trace
    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Manage an existing LeftHand snapshot.

        existing_ref is a dictionary of the form:
        {'source-name': <name of the snapshot>}
        """
        # Check API Version
        self._check_api_version()

        # Potential parent volume for the snapshot
        volume = snapshot['volume']

        if volume.get('replication_status') == 'failed-over':
            err = (_("Managing of snapshots to failed-over volumes is "
                     "not allowed."))
            raise exception.InvalidInput(reason=err)

        target_snap_name = self._get_existing_volume_ref_name(existing_ref)

        # Check for the existence of the virtual volume.
        client = self._login()
        try:
            updates = self._manage_snapshot(client,
                                            volume,
                                            snapshot,
                                            target_snap_name,
                                            existing_ref)
        finally:
            self._logout(client)

        # Return display name to update the name displayed in the GUI and
        # any model updates from retype.
        return updates

    def _manage_snapshot(self, client, volume, snapshot, target_snap_name,
                         existing_ref):
        # Check for the existence of the virtual volume.
        try:
            snapshot_info = client.getSnapshotByName(target_snap_name)
        except hpeexceptions.HTTPNotFound:
            err = (_("Snapshot '%s' doesn't exist on array.") %
                   target_snap_name)
            LOG.error(err)
            raise exception.InvalidInput(reason=err)

        # Make sure the snapshot is being associated with the correct volume.
        try:
            parent_vol = client.getSnapshotParentVolume(target_snap_name)
        except hpeexceptions.HTTPNotFound:
            err = (_("Could not find the parent volume for Snapshot '%s' on "
                     "array.") % target_snap_name)
            LOG.error(err)
            raise exception.InvalidInput(reason=err)

        parent_vol_name = 'volume-' + snapshot['volume_id']
        if parent_vol_name != parent_vol['name']:
            err = (_("The provided snapshot '%s' is not a snapshot of "
                     "the provided volume.") % target_snap_name)
            LOG.error(err)
            raise exception.InvalidInput(reason=err)

        # Generate the new snapshot information based on the new ID.
        new_snap_name = 'snapshot-' + snapshot['id']

        new_vals = {"name": new_snap_name}

        try:
            # Update the existing snapshot with the new name.
            client.modifySnapshot(snapshot_info['id'], new_vals)
        except hpeexceptions.HTTPServerError:
            err = (_("An error occurred while attempting to modify "
                     "Snapshot '%s'.") % snapshot_info['id'])
            LOG.error(err)

        LOG.info("Snapshot '%(ref)s' renamed to '%(new)s'.",
                 {'ref': existing_ref['source-name'], 'new': new_snap_name})

        display_name = None
        if snapshot['display_name']:
            display_name = snapshot['display_name']

        updates = {'display_name': display_name}

        LOG.info("Snapshot %(disp)s '%(new)s' is now being managed.",
                 {'disp': display_name, 'new': new_snap_name})

        return updates

    @cinder_utils.trace
    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of volume to be managed by manage_existing.

        existing_ref is a dictionary of the form:
        {'source-name': <name of the virtual volume>}
        """
        # Check API version.
        self._check_api_version()

        target_vol_name = self._get_existing_volume_ref_name(existing_ref)

        # Make sure the reference is not in use.
        if re.match('volume-*|snapshot-*', target_vol_name):
            reason = _("Reference must be the volume name of an unmanaged "
                       "virtual volume.")
            raise exception.ManageExistingInvalidReference(
                existing_ref=target_vol_name,
                reason=reason)

        # Check for the existence of the virtual volume.
        client = self._login()
        try:
            volume_info = client.getVolumeByName(target_vol_name)
        except hpeexceptions.HTTPNotFound:
            err = (_("Virtual volume '%s' doesn't exist on array.") %
                   target_vol_name)
            LOG.error(err)
            raise exception.InvalidInput(reason=err)
        finally:
            self._logout(client)

        return int(math.ceil(float(volume_info['size']) / units.Gi))

    @cinder_utils.trace
    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return size of volume to be managed by manage_existing.

        existing_ref is a dictionary of the form:
        {'source-name': <name of the virtual volume>}
        """
        # Check API version.
        self._check_api_version()

        target_snap_name = self._get_existing_volume_ref_name(existing_ref)

        # Make sure the reference is not in use.
        if re.match('volume-*|snapshot-*|unm-*', target_snap_name):
            reason = _("Reference must be the name of an unmanaged "
                       "snapshot.")
            raise exception.ManageExistingInvalidReference(
                existing_ref=target_snap_name,
                reason=reason)

        # Check for the existence of the virtual volume.
        client = self._login()
        try:
            snapshot_info = client.getSnapshotByName(target_snap_name)
        except hpeexceptions.HTTPNotFound:
            err = (_("Snapshot '%s' doesn't exist on array.") %
                   target_snap_name)
            LOG.error(err)
            raise exception.InvalidInput(reason=err)
        finally:
            self._logout(client)

        return int(math.ceil(float(snapshot_info['size']) / units.Gi))

    @cinder_utils.trace
    def unmanage(self, volume):
        """Removes the specified volume from Cinder management."""
        # Check API version.
        self._check_api_version()

        # Rename the volume's name to unm-* format so that it can be
        # easily found later.
        client = self._login()
        try:
            volume_info = client.getVolumeByName(volume['name'])
            new_vol_name = 'unm-' + six.text_type(volume['id'])
            options = {'name': new_vol_name}
            client.modifyVolume(volume_info['id'], options)
        finally:
            self._logout(client)

        LOG.info("Virtual volume %(disp)s '%(vol)s' is no longer managed. "
                 "Volume renamed to '%(new)s'.",
                 {'disp': volume['display_name'],
                  'vol': volume['name'],
                  'new': new_vol_name})

    @cinder_utils.trace
    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Cinder management."""
        # Check API version.
        self._check_api_version()

        # Potential parent volume for the snapshot
        volume = snapshot['volume']

        if volume.get('replication_status') == 'failed-over':
            err = (_("Unmanaging of snapshots from 'failed-over' volumes is "
                     "not allowed."))
            LOG.error(err)
            # TODO(leeantho) Change this exception to Invalid when the volume
            # manager supports handling that.
            raise exception.SnapshotIsBusy(snapshot_name=snapshot['id'])

        # Rename the snapshots's name to ums-* format so that it can be
        # easily found later.
        client = self._login()
        try:
            snapshot_info = client.getSnapshotByName(snapshot['name'])
            new_snap_name = 'ums-' + six.text_type(snapshot['id'])
            options = {'name': new_snap_name}
            client.modifySnapshot(snapshot_info['id'], options)
            LOG.info("Snapshot %(disp)s '%(vol)s' is no longer managed. "
                     "Snapshot renamed to '%(new)s'.",
                     {'disp': snapshot['display_name'],
                      'vol': snapshot['name'],
                      'new': new_snap_name})
        finally:
            self._logout(client)

    def _get_existing_volume_ref_name(self, existing_ref):
        """Returns the volume name of an existing reference.

        Checks if an existing volume reference has a source-name element.
        If source-name is not present an error will be thrown.
        """
        if 'source-name' not in existing_ref:
            reason = _("Reference must contain source-name.")
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=reason)

        return existing_ref['source-name']

    def _check_api_version(self):
        """Checks that the API version is correct."""
        if (self.api_version < MIN_API_VERSION):
            ex_msg = (_('Invalid HPELeftHand API version found: %(found)s. '
                        'Version %(minimum)s or greater required for '
                        'manage/unmanage support.')
                      % {'found': self.api_version,
                         'minimum': MIN_API_VERSION})
            LOG.error(ex_msg)
            raise exception.InvalidInput(reason=ex_msg)

    def _get_volume_type(self, type_id):
        ctxt = context.get_admin_context()
        return volume_types.get_volume_type(ctxt, type_id)

    # v2 replication methods
    @cinder_utils.trace
    def failover_host(self, context, volumes, secondary_id=None, groups=None):
        """Force failover to a secondary replication target."""
        if secondary_id and secondary_id == self.FAILBACK_VALUE:
            volume_update_list = self._replication_failback(volumes)
            target_id = None
        else:
            failover_target = None
            for target in self._replication_targets:
                if target['backend_id'] == secondary_id:
                    failover_target = target
                    break
            if not failover_target:
                msg = _("A valid secondary target MUST be specified in order "
                        "to failover.")
                LOG.error(msg)
                raise exception.InvalidReplicationTarget(reason=msg)

            target_id = failover_target['backend_id']
            volume_update_list = []
            for volume in volumes:
                if self._volume_of_replicated_type(volume):
                    # Try and stop the remote snapshot schedule. If the primary
                    # array is down, we will continue with the failover.
                    client = None
                    try:
                        client = self._login(timeout=30)
                        name = volume['name'] + self.REP_SCHEDULE_SUFFIX + (
                            "_Pri")
                        client.stopRemoteSnapshotSchedule(name)
                    except Exception:
                        LOG.warning("The primary array is currently "
                                    "offline, remote copy has been "
                                    "automatically paused.")
                    finally:
                        self._logout(client)

                    # Update provider location to the new array.
                    cl = None
                    try:
                        cl = self._create_replication_client(failover_target)
                        # Stop snapshot schedule
                        try:
                            name = volume['name'] + (
                                self.REP_SCHEDULE_SUFFIX + "_Rmt")
                            cl.stopRemoteSnapshotSchedule(name)
                        except Exception:
                            pass
                        # Make the volume primary so it can be attached after a
                        # fail-over.
                        cl.makeVolumePrimary(volume['name'])

                        # Update the provider info for a proper fail-over.
                        volume_info = cl.getVolumeByName(volume['name'])
                        prov_location = self._update_provider(
                            volume_info,
                            cluster_vip=failover_target['cluster_vip'])
                        volume_update_list.append(
                            {'volume_id': volume['id'],
                             'updates': {'replication_status': 'failed-over',
                                         'provider_location':
                                         prov_location['provider_location']}})
                    except Exception as ex:
                        LOG.error("There was a problem with the failover "
                                  "(%(error)s) and it was unsuccessful. "
                                  "Volume '%(volume)s will not be available "
                                  "on the failed over target.",
                                  {'error': six.text_type(ex),
                                   'volume': volume['id']})
                        volume_update_list.append(
                            {'volume_id': volume['id'],
                             'updates': {'replication_status': 'error'}})
                    finally:
                        self._destroy_replication_client(cl)
                else:
                    # If the volume is not of replicated type, we need to
                    # force the status into error state so a user knows they
                    # do not have access to the volume.
                    volume_update_list.append(
                        {'volume_id': volume['id'],
                         'updates': {'status': 'error'}})

            self._active_backend_id = target_id

        return target_id, volume_update_list, []

    def _do_replication_setup(self):
        default_san_ssh_port = self.configuration.hpelefthand_ssh_port
        default_ssh_conn_timeout = self.configuration.ssh_conn_timeout
        default_san_private_key = self.configuration.san_private_key

        replication_targets = []
        replication_devices = self.configuration.replication_device
        if replication_devices:
            # We do not want to fail if we cannot log into the client here
            # as a failover can still occur, so we need out replication
            # devices to exist.
            for dev in replication_devices:
                remote_array = dict(dev.items())
                # Override and set defaults for certain entries
                remote_array['managed_backend_name'] = (
                    dev.get('managed_backend_name'))
                remote_array['hpelefthand_ssh_port'] = (
                    dev.get('hpelefthand_ssh_port', default_san_ssh_port))
                remote_array['ssh_conn_timeout'] = (
                    dev.get('ssh_conn_timeout', default_ssh_conn_timeout))
                remote_array['san_private_key'] = (
                    dev.get('san_private_key', default_san_private_key))
                # Format hpe3par_iscsi_chap_enabled as a bool
                remote_array['hpelefthand_iscsi_chap_enabled'] = (
                    dev.get('hpelefthand_iscsi_chap_enabled') == 'True')
                remote_array['cluster_id'] = None
                remote_array['cluster_vip'] = None
                array_name = remote_array['backend_id']

                # Make sure we can log into the array, that it has been
                # correctly configured, and its API version meets the
                # minimum requirement.
                cl = None
                try:
                    cl = self._create_replication_client(remote_array)
                    api_version = cl.getApiVersion()
                    cluster_info = cl.getClusterByName(
                        remote_array['hpelefthand_clustername'])
                    remote_array['cluster_id'] = cluster_info['id']
                    virtual_ips = cluster_info['virtualIPAddresses']
                    remote_array['cluster_vip'] = virtual_ips[0]['ipV4Address']

                    if api_version < MIN_API_VERSION:
                        LOG.warning("The secondary array must have an API "
                                    "version of %(min_ver)s or higher. "
                                    "Array '%(target)s' is on %(target_ver)s, "
                                    "therefore it will not be added as a "
                                    "valid replication target.",
                                    {'min_ver': MIN_API_VERSION,
                                     'target': array_name,
                                     'target_ver': api_version})
                    elif not self._is_valid_replication_array(remote_array):
                        LOG.warning("'%s' is not a valid replication array. "
                                    "In order to be valid, backend_id, "
                                    "hpelefthand_api_url, "
                                    "hpelefthand_username, "
                                    "hpelefthand_password, and "
                                    "hpelefthand_clustername, "
                                    "must be specified. If the target is "
                                    "managed, managed_backend_name must be "
                                    "set as well.", array_name)
                    else:
                        replication_targets.append(remote_array)
                except Exception:
                    LOG.error("Could not log in to LeftHand array (%s) with "
                              "the provided credentials.", array_name)
                finally:
                    self._destroy_replication_client(cl)

            self._replication_targets = replication_targets
            if self._is_replication_configured_correct():
                self._replication_enabled = True

    def _replication_failback(self, volumes):
        array_config = {'hpelefthand_api_url':
                        self.configuration.hpelefthand_api_url,
                        'hpelefthand_username':
                        self.configuration.hpelefthand_username,
                        'hpelefthand_password':
                        self.configuration.hpelefthand_password,
                        'hpelefthand_ssh_port':
                        self.configuration.hpelefthand_ssh_port}

        # Make sure the proper steps on the backend have been completed before
        # we allow a failback.
        if not self._is_host_ready_for_failback(volumes, array_config):
            msg = _("The host is not ready to be failed back. Please "
                    "resynchronize the volumes and resume replication on the "
                    "LeftHand backends.")
            LOG.error(msg)
            raise exception.InvalidReplicationTarget(reason=msg)

        cl = None
        volume_update_list = []
        for volume in volumes:
            if self._volume_of_replicated_type(volume):
                try:
                    cl = self._create_replication_client(array_config)
                    # Update the provider info for a proper fail-back.
                    volume_info = cl.getVolumeByName(volume['name'])
                    cluster_info = cl.getClusterByName(
                        self.configuration.hpelefthand_clustername)
                    virtual_ips = cluster_info['virtualIPAddresses']
                    cluster_vip = virtual_ips[0]['ipV4Address']
                    provider_location = self._update_provider(
                        volume_info, cluster_vip=cluster_vip)
                    volume_update_list.append(
                        {'volume_id': volume['id'],
                         'updates': {'replication_status': 'available',
                                     'provider_location':
                                     provider_location['provider_location']}})
                except Exception as ex:
                    # The secondary array was not able to execute the fail-back
                    # properly. The replication status is now in an unknown
                    # state, so we will treat it as an error.
                    LOG.error("There was a problem with the failover "
                              "(%(error)s) and it was unsuccessful. "
                              "Volume '%(volume)s will not be available "
                              "on the failed over target.",
                              {'error': ex,
                               'volume': volume['id']})
                    volume_update_list.append(
                        {'volume_id': volume['id'],
                         'updates': {'replication_status': 'error'}})
                finally:
                    self._destroy_replication_client(cl)
            else:
                # Upon failing back, we can move the non-replicated volumes
                # back into available state.
                volume_update_list.append(
                    {'volume_id': volume['id'],
                     'updates': {'status': 'available'}})

        return volume_update_list

    def _is_host_ready_for_failback(self, volumes, array_config):
        """Checks to make sure the volumes have been synchronized

        This entails ensuring the remote snapshot schedule has been resumed
        on the backends and the secondary volume's data has been copied back
        to the primary.
        """
        is_ready = True
        cl = None
        try:
            for volume in volumes:
                if self._volume_of_replicated_type(volume):
                    schedule_name = volume['name'] + (
                        self.REP_SCHEDULE_SUFFIX + "_Pri")
                    cl = self._create_replication_client(array_config)
                    schedule = cl.getRemoteSnapshotSchedule(schedule_name)
                    schedule = ''.join(schedule)
                    # We need to check the status of the schedule to make sure
                    # it is not paused.
                    result = re.search(r".*paused\s+(\w+)", schedule)
                    is_schedule_active = result.group(1) == 'false'

                    volume_info = cl.getVolumeByName(volume['name'])
                    if not volume_info['isPrimary'] or not is_schedule_active:
                        is_ready = False
                        break
        except Exception as ex:
            LOG.error("There was a problem when trying to determine if "
                      "the volume can be failed-back: %s", ex)
            is_ready = False
        finally:
            self._destroy_replication_client(cl)

        return is_ready

    def _get_replication_targets(self):
        replication_targets = []
        for target in self._replication_targets:
            replication_targets.append(target['backend_id'])

        return replication_targets

    def _is_valid_replication_array(self, target):
        required_flags = ['hpelefthand_api_url', 'hpelefthand_username',
                          'hpelefthand_password', 'backend_id',
                          'hpelefthand_clustername']
        try:
            self.check_replication_flags(target, required_flags)
            return True
        except Exception:
            return False

    def _is_replication_configured_correct(self):
        rep_flag = True
        # Make sure there is at least one replication target.
        if len(self._replication_targets) < 1:
            LOG.error("There must be at least one valid replication "
                      "device configured.")
            rep_flag = False
        return rep_flag

    def _volume_of_replicated_type(self, volume, vol_type_id=None):
        # TODO(kushal) : we will use volume.volume_types when we re-write
        # the design for unit tests to use objects instead of dicts.
        replicated_type = False
        volume_type_id = vol_type_id if vol_type_id else volume.get(
            'volume_type_id')
        if volume_type_id:
            volume_type = self._get_volume_type(volume_type_id)

            extra_specs = volume_type.get('extra_specs')
            if extra_specs and 'replication_enabled' in extra_specs:
                rep_val = extra_specs['replication_enabled']
                replicated_type = (rep_val == "<is> True")

        return replicated_type

    def _does_snapshot_schedule_exist(self, schedule_name, client):
        try:
            exists = client.doesRemoteSnapshotScheduleExist(schedule_name)
        except Exception:
            exists = False
        return exists

    def _get_lefthand_config(self):
        conf = None
        for target in self._replication_targets:
            if target['backend_id'] == self._active_backend_id:
                conf = target
                break

        return conf

    def _do_volume_replication_setup(self, volume, client, optional=None):
        """This function will do or ensure the following:

        -Create volume on main array (already done in create_volume)
        -Create volume on secondary array
        -Make volume remote on secondary array
        -Create the snapshot schedule

        If anything here fails, we will need to clean everything up in
        reverse order, including the original volume.
        """
        schedule_name = volume['name'] + self.REP_SCHEDULE_SUFFIX
        # If there is already a snapshot schedule, the volume is setup
        # for replication on the backend. Start the schedule and return
        # success.
        if self._does_snapshot_schedule_exist(schedule_name + "_Pri", client):
            try:
                client.startRemoteSnapshotSchedule(schedule_name + "_Pri")
            except Exception:
                pass
            return True

        # Grab the extra_spec entries for replication and make sure they
        # are set correctly.
        volume_type = self._get_volume_type(volume["volume_type_id"])
        extra_specs = volume_type.get("extra_specs")

        # Get and check replication sync period
        replication_sync_period = extra_specs.get(
            self.EXTRA_SPEC_REP_SYNC_PERIOD)
        if replication_sync_period:
            replication_sync_period = int(replication_sync_period)
            if replication_sync_period < self.MIN_REP_SYNC_PERIOD:
                msg = (_("The replication sync period must be at least %s "
                         "seconds.") % self.MIN_REP_SYNC_PERIOD)
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
        else:
            # If there is no extra_spec value for replication sync period, we
            # will default it to the required minimum and log a warning.
            replication_sync_period = self.MIN_REP_SYNC_PERIOD
            LOG.warning("There was no extra_spec value for %(spec_name)s, "
                        "so the default value of %(def_val)s will be "
                        "used. To overwrite this, set this value in the "
                        "volume type extra_specs.",
                        {'spec_name': self.EXTRA_SPEC_REP_SYNC_PERIOD,
                         'def_val': self.MIN_REP_SYNC_PERIOD})

        # Get and check retention count
        retention_count = extra_specs.get(
            self.EXTRA_SPEC_REP_RETENTION_COUNT)
        if retention_count:
            retention_count = int(retention_count)
            if retention_count > self.MAX_RETENTION_COUNT:
                msg = (_("The retention count must be %s or less.") %
                       self.MAX_RETENTION_COUNT)
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
        else:
            # If there is no extra_spec value for retention count, we
            # will default it and log a warning.
            retention_count = self.DEFAULT_RETENTION_COUNT
            LOG.warning("There was no extra_spec value for %(spec_name)s, "
                        "so the default value of %(def_val)s will be "
                        "used. To overwrite this, set this value in the "
                        "volume type extra_specs.",
                        {'spec_name': self.EXTRA_SPEC_REP_RETENTION_COUNT,
                         'def_val': self.DEFAULT_RETENTION_COUNT})

        # Get and checkout remote retention count
        remote_retention_count = extra_specs.get(
            self.EXTRA_SPEC_REP_REMOTE_RETENTION_COUNT)
        if remote_retention_count:
            remote_retention_count = int(remote_retention_count)
            if remote_retention_count > self.MAX_REMOTE_RETENTION_COUNT:
                msg = (_("The remote retention count must be %s or less.") %
                       self.MAX_REMOTE_RETENTION_COUNT)
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
        else:
            # If there is no extra_spec value for remote retention count, we
            # will default it and log a warning.
            remote_retention_count = self.DEFAULT_REMOTE_RETENTION_COUNT
            spec_name = self.EXTRA_SPEC_REP_REMOTE_RETENTION_COUNT
            LOG.warning("There was no extra_spec value for %(spec_name)s, "
                        "so the default value of %(def_val)s will be "
                        "used. To overwrite this, set this value in the "
                        "volume type extra_specs.",
                        {'spec_name': spec_name,
                         'def_val': self.DEFAULT_REMOTE_RETENTION_COUNT})

        cl = None
        try:
            # Create volume on secondary system
            for remote_target in self._replication_targets:
                cl = self._create_replication_client(remote_target)

                if optional:
                    optional['clusterName'] = (
                        remote_target['hpelefthand_clustername'])
                cl.createVolume(volume['name'],
                                remote_target['cluster_id'],
                                volume['size'] * units.Gi,
                                optional)

                # Make secondary volume a remote volume
                # NOTE: The snapshot created when making a volume remote is
                # not managed by cinder. This snapshot will be removed when
                # _do_volume_replication_destroy is called.
                snap_name = volume['name'] + self.REP_SNAPSHOT_SUFFIX
                cl.makeVolumeRemote(volume['name'], snap_name)

                # A remote IP address is needed from the cluster in order to
                # create the snapshot schedule.
                remote_ip = cl.getIPFromCluster(
                    remote_target['hpelefthand_clustername'])

                # Destroy remote client
                self._destroy_replication_client(cl)

                # Create remote snapshot schedule on the primary system.
                # We want to start the remote snapshot schedule instantly; a
                # date in the past will do that. We will use the Linux epoch
                # date formatted to ISO 8601 (YYYY-MM-DDTHH:MM:SSZ).
                start_date = "1970-01-01T00:00:00Z"
                remote_vol_name = volume['name']

                client.createRemoteSnapshotSchedule(
                    volume['name'],
                    schedule_name,
                    replication_sync_period,
                    start_date,
                    retention_count,
                    remote_target['hpelefthand_clustername'],
                    remote_retention_count,
                    remote_vol_name,
                    remote_ip,
                    remote_target['hpelefthand_username'],
                    remote_target['hpelefthand_password'])

            return True
        except Exception as ex:
            # Destroy the replication client that was created
            self._destroy_replication_client(cl)
            # Deconstruct what we tried to create
            self._do_volume_replication_destroy(volume, client)
            msg = (_("There was an error setting up a remote schedule "
                     "on the LeftHand arrays: ('%s'). The volume will not be "
                     "recognized as replication type.") %
                   six.text_type(ex))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def _do_volume_replication_destroy(self, volume, client):
        """This will remove all dependencies of a replicated volume

        It should be used when deleting a replication enabled volume
        or if setting up a remote copy group fails. It will try and do the
        following:
        -Delete the snapshot schedule
        -Delete volume and snapshots on secondary array
        -Delete volume and snapshots on primary array
        """
        # Delete snapshot schedule
        try:
            schedule_name = volume['name'] + self.REP_SCHEDULE_SUFFIX
            client.deleteRemoteSnapshotSchedule(schedule_name)
        except Exception:
            pass

        # Delete volume on secondary array(s)
        remote_vol_name = volume['name']
        for remote_target in self._replication_targets:
            try:
                cl = self._create_replication_client(remote_target)
                volume_info = cl.getVolumeByName(remote_vol_name)
                cl.deleteVolume(volume_info['id'])
            except Exception:
                pass
            finally:
                # Destroy the replication client that was created
                self._destroy_replication_client(cl)

        # Delete volume on primary array
        try:
            volume_info = client.getVolumeByName(volume['name'])
            client.deleteVolume(volume_info['id'])
        except Exception:
            pass



# Copyright 2013 Huawei Technologies Co., Ltd
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for cinder.api.contrib.quota_classes.py
"""


import mock

from cinder.api.contrib import quota_classes
from cinder import context
from cinder import db
from cinder import exception
from cinder import quota
from cinder import test
from cinder.tests.unit import fake_constants as fake
from cinder.volume import volume_types


QUOTAS = quota.QUOTAS
GROUP_QUOTAS = quota.GROUP_QUOTAS


def make_body(root=True, gigabytes=1000, snapshots=10,
              volumes=10, backups=10,
              backup_gigabytes=1000, per_volume_gigabytes=-1,
              volume_types_faked=None,
              tenant_id=fake.PROJECT_ID, groups=10):
    resources = {'gigabytes': gigabytes,
                 'snapshots': snapshots,
                 'volumes': volumes,
                 'backups': backups,
                 'per_volume_gigabytes': per_volume_gigabytes,
                 'backup_gigabytes': backup_gigabytes,
                 'groups': groups}
    if not volume_types_faked:
        volume_types_faked = {'fake_type': None}
    for volume_type in volume_types_faked:
        resources['gigabytes_' + volume_type] = -1
        resources['snapshots_' + volume_type] = -1
        resources['volumes_' + volume_type] = -1

    # need to consider preexisting volume types as well
    volume_types = db.volume_type_get_all(context.get_admin_context())

    for volume_type in volume_types:
        resources['gigabytes_' + volume_type] = -1
        resources['snapshots_' + volume_type] = -1
        resources['volumes_' + volume_type] = -1

    if tenant_id:
        resources['id'] = tenant_id
    if root:
        result = {'quota_class_set': resources}
    else:
        result = resources
    return result


def make_response_body(root=True, ctxt=None, quota_class='foo',
                       request_body=None, tenant_id=fake.PROJECT_ID):
    resources = {}
    if not ctxt:
        ctxt = context.get_admin_context()
    resources.update(QUOTAS.get_class_quotas(ctxt, quota_class))
    resources.update(GROUP_QUOTAS.get_class_quotas(ctxt, quota_class))
    if not request_body and not request_body['quota_class_set']:
        resources.update(request_body['quota_class_set'])

    if tenant_id:
        resources['id'] = tenant_id
    if root:
        result = {'quota_class_set': resources}
    else:
        result = resources
    return result


class QuotaClassSetsControllerTest(test.TestCase):

    def setUp(self):
        super(QuotaClassSetsControllerTest, self).setUp()
        self.controller = quota_classes.QuotaClassSetsController()

        self.ctxt = context.get_admin_context()
        self.req = mock.Mock()
        self.req.environ = {'cinder.context': self.ctxt}
        self.req.environ['cinder.context'].is_admin = True

    def test_show(self):
        volume_types.create(self.ctxt, 'fake_type')
        result = self.controller.show(self.req, fake.PROJECT_ID)
        self.assertDictEqual(make_body(), result)

    def test_show_not_authorized(self):
        self.req.environ['cinder.context'].is_admin = False
        self.req.environ['cinder.context'].user_id = fake.USER_ID
        self.req.environ['cinder.context'].project_id = fake.PROJECT_ID
        self.assertRaises(exception.PolicyNotAuthorized, self.controller.show,
                          self.req, fake.PROJECT_ID)

    def test_update(self):
        volume_types.create(self.ctxt, 'fake_type')
        body = make_body(gigabytes=2000, snapshots=15,
                         volumes=5, tenant_id=None)
        result = self.controller.update(self.req, fake.PROJECT_ID, body=body)
        self.assertDictEqual(body, result)

    @mock.patch('cinder.api.openstack.wsgi.Controller.validate_string_length')
    def test_update_limit(self, mock_validate):
        volume_types.create(self.ctxt, 'fake_type')
        body = make_body(volumes=5, tenant_id=None)
        result = self.controller.update(self.req, fake.PROJECT_ID, body=body)
        self.assertEqual(5, result['quota_class_set']['volumes'])
        self.assertTrue(mock_validate.called)

    def test_update_wrong_key(self):
        volume_types.create(self.ctxt, 'fake_type')
        body = {'quota_class_set': {'bad': 100}}
        self.assertRaises(exception.InvalidInput, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_invalid_key_value(self):
        body = {'quota_class_set': {'gigabytes': "should_be_int"}}
        self.assertRaises(exception.ValidationError, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_bad_quota_limit(self):
        body = {'quota_class_set': {'gigabytes': -1000}}
        self.assertRaises(exception.ValidationError, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_no_admin(self):
        self.req.environ['cinder.context'].is_admin = False
        volume_types.create(self.ctxt, 'fake_type')
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller.update, self.req, fake.PROJECT_ID,
                          body=make_body(tenant_id=None))

    def test_update_with_more_volume_types(self):
        volume_types.create(self.ctxt, 'fake_type_1')
        volume_types.create(self.ctxt, 'fake_type_2')
        body = {'quota_class_set': {'gigabytes_fake_type_1': 1111,
                                    'volumes_fake_type_2': 2222}}
        result = self.controller.update(self.req, fake.PROJECT_ID, body=body)
        self.assertDictEqual(make_response_body(ctxt=self.ctxt,
                                                quota_class=fake.PROJECT_ID,
                                                request_body=body,
                                                tenant_id=None),
                             result)



# Copyright (c) 2012 Intel
# Copyright (c) 2012 OpenStack Foundation
# Copyright (c) 2015 EMC Corporation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import math

from oslo_log import log as logging

from cinder.scheduler import filters


LOG = logging.getLogger(__name__)


class CapacityFilter(filters.BaseBackendFilter):
    """Capacity filters based on volume backend's capacity utilization."""

    def backend_passes(self, backend_state, filter_properties):
        """Return True if host has sufficient capacity."""

        volid = None
        # If the volume already exists on this host, don't fail it for
        # insufficient capacity (e.g., if we are retyping)
        if backend_state.backend_id == filter_properties.get('vol_exists_on'):
            return True

        spec = filter_properties.get('request_spec')
        if spec:
            volid = spec.get('volume_id')

        grouping = 'cluster' if backend_state.cluster_name else 'host'
        if filter_properties.get('new_size'):
            # If new_size is passed, we are allocating space to extend a volume
            requested_size = (int(filter_properties.get('new_size')) -
                              int(filter_properties.get('size')))
            LOG.debug('Checking if %(grouping)s %(grouping_name)s can extend '
                      'the volume %(id)s in %(size)s GB',
                      {'grouping': grouping,
                       'grouping_name': backend_state.backend_id, 'id': volid,
                       'size': requested_size})
        else:
            requested_size = filter_properties.get('size')
            LOG.debug('Checking if %(grouping)s %(grouping_name)s can create '
                      'a %(size)s GB volume (%(id)s)',
                      {'grouping': grouping,
                       'grouping_name': backend_state.backend_id, 'id': volid,
                       'size': requested_size})

        # requested_size is 0 means that it's a manage request.
        if requested_size == 0:
            return True

        if backend_state.free_capacity_gb is None:
            # Fail Safe
            LOG.error("Free capacity not set: "
                      "volume node info collection broken.")
            return False

        free_space = backend_state.free_capacity_gb
        total_space = backend_state.total_capacity_gb
        reserved = float(backend_state.reserved_percentage) / 100
        if free_space in ['infinite', 'unknown']:
            # NOTE(zhiteng) for those back-ends cannot report actual
            # available capacity, we assume it is able to serve the
            # request.  Even if it was not, the retry mechanism is
            # able to handle the failure by rescheduling
            return True
        elif total_space in ['infinite', 'unknown']:
            # If total_space is 'infinite' or 'unknown' and reserved
            # is 0, we assume the back-ends can serve the request.
            # If total_space is 'infinite' or 'unknown' and reserved
            # is not 0, we cannot calculate the reserved space.
            # float(total_space) will throw an exception. total*reserved
            # also won't work. So the back-ends cannot serve the request.
            if reserved == 0:
                return True
            LOG.debug("Cannot calculate GB of reserved space (%s%%) with "
                      "backend's reported total capacity '%s'",
                      backend_state.reserved_percentage, total_space)
            return False
        total = float(total_space)
        if total <= 0:
            LOG.warning("Insufficient free space for volume creation. "
                        "Total capacity is %(total).2f on %(grouping)s "
                        "%(grouping_name)s.",
                        {"total": total,
                         "grouping": grouping,
                         "grouping_name": backend_state.backend_id})
            return False

        # Calculate how much free space is left after taking into account
        # the reserved space.
        free = free_space - math.floor(total * reserved)

        # NOTE(xyang): If 'provisioning:type' is 'thick' in extra_specs,
        # we will not use max_over_subscription_ratio and
        # provisioned_capacity_gb to determine whether a volume can be
        # provisioned. Instead free capacity will be used to evaluate.
        thin = True
        vol_type = filter_properties.get('volume_type', {}) or {}
        provision_type = vol_type.get('extra_specs', {}).get(
            'provisioning:type')
        if provision_type == 'thick':
            thin = False

        msg_args = {"grouping_name": backend_state.backend_id,
                    "grouping": grouping,
                    "requested": requested_size,
                    "available": free}
        # Only evaluate using max_over_subscription_ratio if
        # thin_provisioning_support is True. Check if the ratio of
        # provisioned capacity over total capacity has exceeded over
        # subscription ratio.
        if (thin and backend_state.thin_provisioning_support and
                backend_state.max_over_subscription_ratio >= 1):
            provisioned_ratio = ((backend_state.provisioned_capacity_gb +
                                  requested_size) / total)
            LOG.debug("Checking provisioning for request of %s GB. "
                      "Backend: %s", requested_size, backend_state)
            if provisioned_ratio > backend_state.max_over_subscription_ratio:
                msg_args = {
                    "provisioned_ratio": provisioned_ratio,
                    "oversub_ratio": backend_state.max_over_subscription_ratio,
                    "grouping": grouping,
                    "grouping_name": backend_state.backend_id,
                }
                LOG.warning(
                    "Insufficient free space for thin provisioning. "
                    "The ratio of provisioned capacity over total capacity "
                    "%(provisioned_ratio).2f has exceeded the maximum over "
                    "subscription ratio %(oversub_ratio).2f on %(grouping)s "
                    "%(grouping_name)s.", msg_args)
                return False
            else:
                # Thin provisioning is enabled and projected over-subscription
                # ratio does not exceed max_over_subscription_ratio. The host
                # passes if "adjusted" free virtual capacity is enough to
                # accommodate the volume. Adjusted free virtual capacity is
                # the currently available free capacity (taking into account
                # of reserved space) which we can over-subscribe.
                adjusted_free_virtual = (
                    free * backend_state.max_over_subscription_ratio)
                res = adjusted_free_virtual >= requested_size
                if not res:
                    msg_args = {"available": adjusted_free_virtual,
                                "size": requested_size,
                                "grouping": grouping,
                                "grouping_name": backend_state.backend_id}
                    LOG.warning("Insufficient free virtual space "
                                "(%(available)sGB) to accommodate thin "
                                "provisioned %(size)sGB volume on %(grouping)s"
                                " %(grouping_name)s.", msg_args)
                else:
                    LOG.debug("Space information for volume creation "
                              "on %(grouping)s %(grouping_name)s "
                              "(requested / avail): "
                              "%(requested)s/%(available)s", msg_args)
                return res
        elif thin and backend_state.thin_provisioning_support:
            LOG.warning("Filtering out %(grouping)s %(grouping_name)s "
                        "with an invalid maximum over subscription ratio "
                        "of %(oversub_ratio).2f. The ratio should be a "
                        "minimum of 1.0.",
                        {"oversub_ratio":
                            backend_state.max_over_subscription_ratio,
                         "grouping": grouping,
                         "grouping_name": backend_state.backend_id})
            return False

        if free < requested_size:
            LOG.warning("Insufficient free space for volume creation "
                        "on %(grouping)s %(grouping_name)s (requested / "
                        "avail): %(requested)s/%(available)s",
                        msg_args)
            return False

        LOG.debug("Space information for volume creation "
                  "on %(grouping)s %(grouping_name)s (requested / avail): "
                  "%(requested)s/%(available)s", msg_args)

        return True



# Copyright (c) 2016 Synology Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the Synology iSCSI volume driver."""

import mock

from cinder import exception
from cinder import test
from cinder.tests.unit import fake_constants as fake
from cinder.volume import configuration as conf
from cinder.volume.drivers.synology import synology_common as common
from cinder.volume.drivers.synology import synology_iscsi

VOLUME_ID = fake.VOLUME_ID
TARGET_NAME_PREFIX = 'Cinder-Target-'
IP = '10.10.10.10'
IQN = 'iqn.2000-01.com.synology:' + TARGET_NAME_PREFIX + VOLUME_ID
TRG_ID = 1
VOLUME = {
    'name': fake.VOLUME_NAME,
    'id': VOLUME_ID,
    'display_name': 'fake_volume',
    'size': 10,
    'provider_location': '%s:3260,%d %s 1' % (IP, TRG_ID, IQN),
}
NEW_VOLUME_ID = fake.VOLUME2_ID
IQN2 = 'iqn.2000-01.com.synology:' + TARGET_NAME_PREFIX + NEW_VOLUME_ID
NEW_TRG_ID = 2
NEW_VOLUME = {
    'name': fake.VOLUME2_NAME,
    'id': NEW_VOLUME_ID,
    'display_name': 'new_fake_volume',
    'size': 10,
    'provider_location': '%s:3260,%d %s 1' % (IP, NEW_TRG_ID, IQN2),
}
SNAPSHOT_ID = fake.SNAPSHOT_ID
SNAPSHOT = {
    'name': fake.SNAPSHOT_NAME,
    'id': SNAPSHOT_ID,
    'volume_id': VOLUME_ID,
    'volume_name': VOLUME['name'],
    'volume_size': 10,
    'display_name': 'fake_snapshot',
}
DS_SNAPSHOT_UUID = 'ca86a56a-40d8-4210-974c-ef15dbf01cba'
SNAPSHOT_METADATA = {
    'metadata': {
        'ds_snapshot_UUID': DS_SNAPSHOT_UUID
    }
}
INITIATOR_IQN = 'iqn.1993-08.org.debian:01:604af6a341'
CONNECTOR = {
    'initiator': INITIATOR_IQN,
}
CONTEXT = {
}
LOCAL_PATH = '/dev/isda'
IMAGE_SERVICE = 'image_service'
IMAGE_ID = 1
IMAGE_META = {
    'id': IMAGE_ID
}
NODE_UUID = '72003c93-2db2-4f00-a169-67c5eae86bb1'
HOST = {
}


class SynoISCSIDriverTestCase(test.TestCase):

    @mock.patch.object(common.SynoCommon,
                       '_get_node_uuid',
                       return_value=NODE_UUID)
    @mock.patch.object(common, 'APIRequest')
    def setUp(self, _request, _get_node_uuid):
        super(SynoISCSIDriverTestCase, self).setUp()

        self.conf = self.setup_configuration()
        self.driver = synology_iscsi.SynoISCSIDriver(configuration=self.conf)
        self.driver.common = common.SynoCommon(self.conf, 'iscsi')

    def setup_configuration(self):
        config = mock.Mock(spec=conf.Configuration)
        config.use_chap_auth = False
        config.target_protocol = 'iscsi'
        config.target_ip_address = IP
        config.synology_admin_port = 5000
        config.synology_username = 'admin'
        config.synology_password = 'admin'
        config.synology_ssl_verify = True
        config.synology_one_time_pass = '123456'
        config.volume_dd_blocksize = 1

        return config

    def test_check_for_setup_error(self):
        self.driver.common.check_for_setup_error = mock.Mock()

        result = self.driver.check_for_setup_error()

        self.driver.common.check_for_setup_error.assert_called_with()
        self.assertIsNone(result)

    def test_create_volume(self):
        self.driver.common.create_volume = mock.Mock()

        result = self.driver.create_volume(VOLUME)

        self.driver.common.create_volume.assert_called_with(VOLUME)
        self.assertIsNone(result)

    def test_delete_volume(self):
        self.driver.common.delete_volume = mock.Mock()

        result = self.driver.delete_volume(VOLUME)

        self.driver.common.delete_volume.assert_called_with(VOLUME)
        self.assertIsNone(result)

    def test_create_cloned_volume(self):
        self.driver.common.create_cloned_volume = mock.Mock()

        result = self.driver.create_cloned_volume(VOLUME, NEW_VOLUME)

        self.driver.common.create_cloned_volume.assert_called_with(
            VOLUME, NEW_VOLUME)
        self.assertIsNone(result)

    def test_extend_volume(self):
        new_size = 20

        self.driver.common.extend_volume = mock.Mock()

        result = self.driver.extend_volume(VOLUME, new_size)

        self.driver.common.extend_volume.assert_called_with(
            VOLUME, new_size)
        self.assertIsNone(result)

    def test_extend_volume_wrong_size(self):
        wrong_new_size = 1

        self.driver.common.extend_volume = mock.Mock()

        result = self.driver.extend_volume(VOLUME, wrong_new_size)

        self.driver.common.extend_volume.assert_not_called()
        self.assertIsNone(result)

    def test_create_volume_from_snapshot(self):
        self.driver.common.create_volume_from_snapshot = mock.Mock()

        result = self.driver.create_volume_from_snapshot(VOLUME, SNAPSHOT)

        (self.driver.common.
            create_volume_from_snapshot.assert_called_with(VOLUME, SNAPSHOT))
        self.assertIsNone(result)

    def test_update_migrated_volume(self):
        fake_ret = {'_name_id': VOLUME['id']}
        status = ''
        self.driver.common.update_migrated_volume = (
            mock.Mock(return_value=fake_ret))

        result = self.driver.update_migrated_volume(CONTEXT,
                                                    VOLUME,
                                                    NEW_VOLUME,
                                                    status)

        (self.driver.common.update_migrated_volume.
            assert_called_with(VOLUME, NEW_VOLUME))
        self.assertEqual(fake_ret, result)

    def test_create_snapshot(self):
        self.driver.common.create_snapshot = (
            mock.Mock(return_value=SNAPSHOT_METADATA))

        result = self.driver.create_snapshot(SNAPSHOT)

        self.driver.common.create_snapshot.assert_called_with(SNAPSHOT)
        self.assertDictEqual(SNAPSHOT_METADATA, result)

    def test_delete_snapshot(self):
        self.driver.common.delete_snapshot = mock.Mock()

        result = self.driver.delete_snapshot(SNAPSHOT)

        self.driver.common.delete_snapshot.assert_called_with(SNAPSHOT)
        self.assertIsNone(result)

    def test_get_volume_stats(self):
        self.driver.common.update_volume_stats = mock.MagicMock()

        result = self.driver.get_volume_stats(True)

        self.driver.common.update_volume_stats.assert_called_with()
        self.assertEqual(self.driver.stats, result)

        result = self.driver.get_volume_stats(False)

        self.driver.common.update_volume_stats.assert_called_with()
        self.assertEqual(self.driver.stats, result)

    def test_get_volume_stats_error(self):
        self.driver.common.update_volume_stats = (
            mock.MagicMock(side_effect=exception.VolumeDriverException(
                message='dont care')))

        self.assertRaises(exception.VolumeDriverException,
                          self.driver.get_volume_stats,
                          True)

    def test_create_export(self):
        provider_auth = 'CHAP username password'
        provider_location = '%s:3260,%d %s 1' % (IP, TRG_ID, IQN)

        self.driver.common.is_lun_mapped = mock.Mock(return_value=False)
        self.driver.common.create_iscsi_export = (
            mock.Mock(return_value=(IQN, TRG_ID, provider_auth)))
        self.driver.common.get_provider_location = (
            mock.Mock(return_value=provider_location))

        result = self.driver.create_export(CONTEXT, VOLUME, CONNECTOR)

        self.driver.common.is_lun_mapped.assert_called_with(VOLUME['name'])
        (self.driver.common.create_iscsi_export.
            assert_called_with(VOLUME['name'], VOLUME['id']))
        self.driver.common.get_provider_location.assert_called_with(IQN,
                                                                    TRG_ID)
        self.assertEqual(provider_location, result['provider_location'])
        self.assertEqual(provider_auth, result['provider_auth'])

    def test_create_export_is_mapped(self):
        self.driver.common.is_lun_mapped = mock.Mock(return_value=True)
        self.driver.common.create_iscsi_export = mock.Mock()
        self.driver.common.get_provider_location = mock.Mock()

        result = self.driver.create_export(CONTEXT, VOLUME, CONNECTOR)

        self.driver.common.is_lun_mapped.assert_called_with(VOLUME['name'])
        self.driver.common.create_iscsi_export.assert_not_called()
        self.driver.common.get_provider_location.assert_not_called()
        self.assertEqual({}, result)

    def test_create_export_error(self):
        provider_location = '%s:3260,%d %s 1' % (IP, TRG_ID, IQN)

        self.driver.common.is_lun_mapped = mock.Mock(return_value=False)
        self.driver.common.create_iscsi_export = (
            mock.Mock(side_effect=exception.InvalidInput(reason='dont care')))
        self.driver.common.get_provider_location = (
            mock.Mock(return_value=provider_location))

        self.assertRaises(exception.ExportFailure,
                          self.driver.create_export,
                          CONTEXT,
                          VOLUME,
                          CONNECTOR)
        self.driver.common.is_lun_mapped.assert_called_with(VOLUME['name'])
        self.driver.common.get_provider_location.assert_not_called()

    def test_remove_export(self):
        self.driver.common.is_lun_mapped = mock.Mock(return_value=True)
        self.driver.common.remove_iscsi_export = mock.Mock()
        self.driver.common.get_iqn_and_trgid = (
            mock.Mock(return_value=('', TRG_ID)))

        _, trg_id = (self.driver.common.
                     get_iqn_and_trgid(VOLUME['provider_location']))
        result = self.driver.remove_export(CONTEXT, VOLUME)

        self.driver.common.is_lun_mapped.assert_called_with(VOLUME['name'])
        (self.driver.common.get_iqn_and_trgid.
            assert_called_with(VOLUME['provider_location']))
        (self.driver.common.remove_iscsi_export.
            assert_called_with(VOLUME['name'], trg_id))
        self.assertIsNone(result)

    def test_remove_export_not_mapped(self):
        self.driver.common.is_lun_mapped = mock.Mock(return_value=False)
        self.driver.common.remove_iscsi_export = mock.Mock()
        self.driver.common.get_iqn_and_trgid = mock.Mock()

        result = self.driver.remove_export(CONTEXT, VOLUME)

        self.driver.common.is_lun_mapped.assert_called_with(VOLUME['name'])
        self.driver.common.get_iqn_and_trgid.assert_not_called()
        self.driver.common.remove_iscsi_export.assert_not_called()
        self.assertIsNone(result)

    def test_remove_export_error(self):
        self.driver.common.is_lun_mapped = mock.Mock(return_value=True)
        self.driver.common.remove_iscsi_export = (
            mock.Mock(side_effect= exception.RemoveExportException(
                volume=VOLUME, reason='dont care')))

        self.assertRaises(exception.RemoveExportException,
                          self.driver.remove_export,
                          CONTEXT,
                          VOLUME)

    def test_remove_export_error_get_lun_mapped(self):
        self.driver.common.remove_iscsi_export = mock.Mock()
        self.driver.common.get_iqn_and_trgid = mock.Mock()
        self.driver.common.is_lun_mapped = (
            mock.Mock(side_effect=common.SynoLUNNotExist(
                message='dont care')))

        result = self.driver.remove_export(CONTEXT, VOLUME)

        self.assertIsNone(result)
        self.driver.common.get_iqn_and_trgid.assert_not_called()
        self.driver.common.remove_iscsi_export.assert_not_called()

    def test_initialize_connection(self):
        iscsi_properties = {
            'target_discovered': False,
            'target_iqn': IQN,
            'target_portal': '%s:3260' % self.conf.target_ip_address,
            'volume_id': VOLUME['id'],
            'access_mode': 'rw',
            'discard': False
        }

        self.driver.common.get_iscsi_properties = (
            mock.Mock(return_value=iscsi_properties))
        self.conf.safe_get = mock.Mock(return_value='iscsi')

        result = self.driver.initialize_connection(VOLUME, CONNECTOR)

        self.driver.common.get_iscsi_properties.assert_called_with(VOLUME)
        self.conf.safe_get.assert_called_with('target_protocol')
        self.assertEqual('iscsi', result['driver_volume_type'])
        self.assertDictEqual(iscsi_properties, result['data'])

    def test_initialize_connection_error(self):
        self.driver.common.get_iscsi_properties = (
            mock.Mock(side_effect=exception.InvalidInput(reason='dont care')))

        self.assertRaises(exception.InvalidInput,
                          self.driver.initialize_connection,
                          VOLUME,
                          CONNECTOR)



#!/usr/bin/env python
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Starter script for Cinder Volume."""

import eventlet
import eventlet.tpool
import os

# Monkey patching must go before the oslo.log import, otherwise
# oslo.context will not use greenthread thread local and all greenthreads
# will share the same context.
if os.name == 'nt':
    # eventlet monkey patching the os module causes subprocess.Popen to fail
    # on Windows when using pipes due to missing non-blocking IO support.
    eventlet.monkey_patch(os=False)
else:
    eventlet.monkey_patch()

import logging as python_logging
import re

from cinder import exception
from cinder import objects


import shlex
import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_privsep import priv_context
from oslo_reports import guru_meditation_report as gmr
from oslo_reports import opts as gmr_opts

from cinder import i18n
i18n.enable_lazy()

# Need to register global_opts
from cinder.common import config  # noqa
from cinder.common import constants
from cinder.db import api as session
from cinder.i18n import _
from cinder import service
from cinder import utils
from cinder import version


CONF = cfg.CONF

host_opt = cfg.StrOpt('backend_host', help='Backend override of host value.')
CONF.register_cli_opt(host_opt)

backend_name_opt = cfg.StrOpt(
    'backend_name',
    help='NOTE: For Windows internal use only. The name of the backend to be '
         'managed by this process. It must be one of the backends defined '
         'using the "enabled_backends" option. Note that normally, this '
         'should not be used directly. Cinder uses it internally in order to '
         'spawn subprocesses on Windows.')
CONF.register_cli_opt(backend_name_opt)


# TODO(geguileo): Once we complete the work on A-A update the option's help.
cluster_opt = cfg.StrOpt('cluster',
                         default=None,
                         help='Name of this cluster. Used to group volume '
                              'hosts that share the same backend '
                              'configurations to work in HA Active-Active '
                              'mode.  Active-Active is not yet supported.')
CONF.register_opt(cluster_opt)

LOG = None

service_started = False


def _launch_service(launcher, backend):
    CONF.register_opt(host_opt, group=backend)
    backend_host = getattr(CONF, backend).backend_host
    host = "%s@%s" % (backend_host or CONF.host, backend)
    # We also want to set cluster to None on empty strings, and we
    # ignore leading and trailing spaces.
    cluster = CONF.cluster and CONF.cluster.strip()
    cluster = (cluster or None) and '%s@%s' % (cluster, backend)
    try:
        server = service.Service.create(host=host,
                                        service_name=backend,
                                        binary=constants.VOLUME_BINARY,
                                        coordination=True,
                                        cluster=cluster)
    except Exception:
        LOG.exception('Volume service %s failed to start.', host)
    else:
        # Dispose of the whole DB connection pool here before
        # starting another process.  Otherwise we run into cases where
        # child processes share DB connections which results in errors.
        session.dispose_engine()
        launcher.launch_service(server)
        _notify_service_started()


def _ensure_service_started():
    if not service_started:
        LOG.error('No volume service(s) started successfully, terminating.')
        sys.exit(1)


def _notify_service_started():
    global service_started
    service_started = True


def _launch_services_win32():
    if CONF.backend_name and CONF.backend_name not in CONF.enabled_backends:
        msg = _('The explicitly passed backend name "%(backend_name)s" is not '
                'among the enabled backends: %(enabled_backends)s.')
        raise exception.InvalidInput(
            reason=msg % dict(backend_name=CONF.backend_name,
                              enabled_backends=CONF.enabled_backends))

    # We'll avoid spawning a subprocess if a single backend is requested.
    single_backend_name = (CONF.enabled_backends[0]
                           if len(CONF.enabled_backends) == 1
                           else CONF.backend_name)
    if single_backend_name:
        launcher = service.get_launcher()
        _launch_service(launcher, single_backend_name)
    elif CONF.enabled_backends:
        # We're using the 'backend_name' argument, requesting a certain backend
        # and constructing the service object within the child process.
        launcher = service.WindowsProcessLauncher()
        py_script_re = re.compile(r'.*\.py\w?$')
        for backend in filter(None, CONF.enabled_backends):
            cmd = sys.argv + ['--backend_name=%s' % backend]
            # Recent setuptools versions will trim '-script.py' and '.exe'
            # extensions from sys.argv[0].
            if py_script_re.match(sys.argv[0]):
                cmd = [sys.executable] + cmd
            launcher.add_process(cmd)
            _notify_service_started()

    _ensure_service_started()

    launcher.wait()


def _launch_services_posix():
    launcher = service.get_launcher()

    for backend in filter(None, CONF.enabled_backends):
        _launch_service(launcher, backend)

    _ensure_service_started()

    launcher.wait()


def main():
    objects.register_all()
    gmr_opts.set_defaults(CONF)
    CONF(sys.argv[1:], project='cinder',
         version=version.version_string())
    logging.setup(CONF, "cinder")
    python_logging.captureWarnings(True)
    priv_context.init(root_helper=shlex.split(utils.get_root_helper()))
    utils.monkey_patch()
    gmr.TextGuruMeditation.setup_autorun(version, conf=CONF)
    global LOG
    LOG = logging.getLogger(__name__)

    if not CONF.enabled_backends:
        LOG.error('Configuration for cinder-volume does not specify '
                  '"enabled_backends". Using DEFAULT section to configure '
                  'drivers is not supported since Ocata.')
        sys.exit(1)

    if os.name == 'nt':
        # We cannot use oslo.service to spawn multiple services on Windows.
        # It relies on forking, which is not available on Windows.
        # Furthermore, service objects are unmarshallable objects that are
        # passed to subprocesses.
        _launch_services_win32()
    else:
        _launch_services_posix()



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six
import time

from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import excutils
from oslo_utils import importutils
from oslo_utils import uuidutils

from cinder import exception
from cinder.i18n import _
from cinder.objects import fields
from cinder.volume.drivers.dell_emc.vnx import common
from cinder.volume.drivers.san.san import san_opts
from cinder.volume import volume_types
from cinder.volume import volume_utils

storops = importutils.try_import('storops')


storops = importutils.try_import('storops')
LOG = logging.getLogger(__name__)


def init_ops(configuration):
    configuration.append_config_values(common.VNX_OPTS)
    configuration.append_config_values(san_opts)


def get_metadata(volume):
    # Since versionedobjects is partially merged, metadata
    # may come from 'volume_metadata' or 'metadata', here
    # we need to take care both of them.
    volume_metadata = {}
    if 'volume_metadata' in volume:
        for metadata in volume['volume_metadata']:
            volume_metadata[metadata['key']] = metadata['value']
        return volume_metadata
    return volume['metadata'] if 'metadata' in volume else {}


def dump_provider_location(location_dict):
    return '|'.join([k + '^' + v for k, v in location_dict.items()])


def build_provider_location(system, lun_type, lun_id, base_lun_name, version):
    """Builds provider_location for volume or snapshot.

    :param system: VNX serial number
    :param lun_id: LUN ID in VNX
    :param lun_type: 'lun' or 'smp'
    :param base_lun_name: primary LUN name,
                          it will be used when creating snap lun
    :param version: driver version
    """
    location_dict = {'system': system,
                     'type': lun_type,
                     'id': six.text_type(lun_id),
                     'base_lun_name': six.text_type(base_lun_name),
                     'version': version}
    return dump_provider_location(location_dict)


def extract_provider_location(provider_location, key):
    """Extracts value of the specified field from provider_location string.

    :param provider_location: provider_location string
    :param key: field name of the value that to be extracted
    :return: value of the specified field if it exists, otherwise,
             None is returned
    """
    if not provider_location:
        return None

    kvps = provider_location.split('|')
    for kvp in kvps:
        fields = kvp.split('^')
        if len(fields) == 2 and fields[0] == key:
            return fields[1]


def update_provider_location(provider_location, items):
    """Updates provider_location with new dict items.

    :param provider_location: volume's provider_location.
    :param items: dict items for updating.
    """
    location_dict = {tp.split('^')[0]: tp.split('^')[1]
                     for tp in provider_location.split('|')}
    for key, value in items.items():
        location_dict[key] = value
    return dump_provider_location(location_dict)


def update_remote_provider_location(volume, client):
    """Update volume provider_location after volume failed-over."""
    provider_location = volume.provider_location
    updated = {}
    updated['system'] = client.get_serial()
    updated['id'] = six.text_type(
        client.get_lun(name=volume.name).lun_id)
    provider_location = update_provider_location(
        provider_location, updated)
    return provider_location


def get_pool_from_host(host):
    return volume_utils.extract_host(host, 'pool')


def wait_until(condition, timeout=None, interval=common.INTERVAL_5_SEC,
               reraise_arbiter=lambda ex: True, *args, **kwargs):
    start_time = time.time()
    if not timeout:
        timeout = common.DEFAULT_TIMEOUT

    def _inner():
        try:
            test_value = condition(*args, **kwargs)
        except Exception as ex:
            test_value = False
            with excutils.save_and_reraise_exception(
                    reraise=reraise_arbiter(ex)):
                LOG.debug('Exception raised when executing %(condition_name)s '
                          'in wait_until. Message: %(msg)s',
                          {'condition_name': condition.__name__,
                           'msg': ex.message})
        if test_value:
            raise loopingcall.LoopingCallDone()

        if int(time.time()) - start_time > timeout:
            msg = (_('Timeout waiting for %(condition_name)s in wait_until.')
                   % {'condition_name': condition.__name__})
            LOG.error(msg)
            raise common.WaitUtilTimeoutException(msg)

    timer = loopingcall.FixedIntervalLoopingCall(_inner)
    timer.start(interval=interval).wait()


def validate_storage_migration(volume, target_host, src_serial, src_protocol):
    if 'location_info' not in target_host['capabilities']:
        LOG.warning("Failed to get pool name and "
                    "serial number. 'location_info' "
                    "from %s.", target_host['host'])
        return False
    info = target_host['capabilities']['location_info']
    LOG.debug("Host for migration is %s.", info)
    try:
        serial_number = info.split('|')[1]
    except AttributeError:
        LOG.warning('Error on getting serial number '
                    'from %s.', target_host['host'])
        return False
    if serial_number != src_serial:
        LOG.debug('Skip storage-assisted migration because '
                  'target and source backend are not managing '
                  'the same array.')
        return False
    if (target_host['capabilities']['storage_protocol'] != src_protocol
            and get_original_status(volume) == 'in-use'):
        LOG.debug('Skip storage-assisted migration because '
                  'in-use volume can not be '
                  'migrate between different protocols.')
        return False
    return True


def retype_need_migration(volume, old_provision, new_provision, host):
    if volume['host'] != host['host']:
        return True

    lun_type = extract_provider_location(volume['provider_location'], 'type')
    if lun_type == 'smp':
        return True

    if old_provision != new_provision:
        if retype_need_turn_on_compression(old_provision, new_provision):
            return False
        else:
            return True

    return False


def retype_need_turn_on_compression(old_provision, new_provision):
    return (old_provision in [storops.VNXProvisionEnum.THIN,
                              storops.VNXProvisionEnum.THICK]
            and new_provision == storops.VNXProvisionEnum.COMPRESSED)


def retype_need_change_tier(old_tier, new_tier):
    return new_tier is not None and old_tier != new_tier


def get_original_status(volume):
    if not volume['volume_attachment']:
        return 'available'
    else:
        return 'in-use'


def construct_snap_name(volume):
    """Return snapshot name."""
    if is_snapcopy_enabled(volume):
        return 'snap-as-vol-' + six.text_type(volume.name_id)
    else:
        return 'tmp-snap-' + six.text_type(volume.name_id)


def construct_mirror_name(volume):
    """Constructs MirrorView name for volume."""
    return 'mirror_' + six.text_type(volume.id)


def construct_group_name(group):
    """Constructs MirrorGroup name for volumes.

    VNX only allows for 32-character group name, so
    trim the dash(-) from group id.
    """
    return group.id.replace('-', '')


def construct_tmp_cg_snap_name(cg_name):
    """Return CG snapshot name."""
    return 'tmp-snap-' + six.text_type(cg_name)


def construct_tmp_lun_name(lun_name):
    """Constructs a time-based temporary LUN name."""
    return '%(src)s-%(ts)s' % {'src': lun_name,
                               'ts': int(time.time())}


def construct_smp_name(snap_id):
    return 'tmp-smp-' + six.text_type(snap_id)


def is_snapcopy_enabled(volume):
    meta = get_metadata(volume)
    return 'snapcopy' in meta and meta['snapcopy'].lower() == 'true'


def is_async_migrate_enabled(volume, default=True):
    extra_specs = common.ExtraSpecs.from_volume(volume)
    if extra_specs.is_replication_enabled:
        # For replication-enabled volume, we should not use the async-cloned
        # volume, or setup replication would fail with
        # VNXMirrorLunNotAvailableError
        return False
    meta = get_metadata(volume)
    if 'async_migrate' not in meta:
        return default
    return 'async_migrate' in meta and meta['async_migrate'].lower() == 'true'


def get_migration_rate(volume):
    metadata = get_metadata(volume)
    rate = metadata.get('migrate_rate', None)
    if rate:
        if rate.lower() in storops.VNXMigrationRate.values():
            return storops.VNXMigrationRate.parse(rate.lower())
        else:
            LOG.warning('Unknown migration rate specified, '
                        'using [high] as migration rate.')

            return storops.VNXMigrationRate.HIGH


def check_type_matched(volume):
    """Check volume type and group type

    This will make sure they do not conflict with each other.

    :param volume: volume to be checked
    :returns: None
    :raises: InvalidInput

    """

    # If volume is not a member of group, skip this check anyway.
    if not volume.group:
        return
    extra_specs = common.ExtraSpecs.from_volume(volume)
    group_specs = common.ExtraSpecs.from_group(volume.group)

    if not (group_specs.is_group_replication_enabled ==
            extra_specs.is_replication_enabled):
        msg = _('Replication should be enabled or disabled for both '
                'volume or group. volume replication status: %(vol_status)s, '
                'group replication status: %(group_status)s') % {
                    'vol_status': extra_specs.is_replication_enabled,
                    'group_status': group_specs.is_group_replication_enabled}
        raise exception.InvalidInput(reason=msg)


def check_rep_status_matched(group):
    """Check replication status for group.

    Group status must be enabled before proceeding.
    """
    group_specs = common.ExtraSpecs.from_group(group)
    if group_specs.is_group_replication_enabled:
        if group.replication_status != fields.ReplicationStatus.ENABLED:
            msg = _('Replication status should be %s for replication-enabled '
                    'group.') % fields.ReplicationStatus.ENABLED
            raise exception.InvalidInput(reason=msg)
    else:
        LOG.info('Replication is not enabled on group %s, skip status check.',
                 group.id)


def update_res_without_poll(res):
    with res.with_no_poll():
        res.update()


def update_res_with_poll(res):
    with res.with_poll():
        res.update()


def get_base_lun_name(volume):
    """Returns base LUN name for LUN/snapcopy LUN."""
    base_name = extract_provider_location(
        volume.provider_location, 'base_lun_name')
    if base_name is None or base_name == 'None':
        return volume.name
    return base_name


def sift_port_white_list(port_white_list, registered_io_ports):
    """Filters out the unregistered ports.

    Goes through the `port_white_list`, and filters out the ones not
    registered (that is not in `registered_io_ports`).
    """
    valid_port_list = []
    LOG.debug('Filter ports in [%(white)s}] but not in [%(reg_ports)s].',
              {'white': ','.join(
                  [port.display_name for port in port_white_list]),
               'reg_ports': ','.join(
                   [port.display_name for port in registered_io_ports])})
    for io_port in port_white_list:
        if io_port not in registered_io_ports:
            LOG.debug('Skipped SP port %(port)s due to it is not registered. '
                      'The registered IO ports: %(reg_ports)s.',
                      {'port': io_port, 'reg_ports': registered_io_ports})
        else:
            valid_port_list.append(io_port)

    return valid_port_list


def convert_to_tgt_list_and_itor_tgt_map(zone_mapping):
    """Function to process data from lookup service.

    :param zone_mapping: mapping is the data from the zone lookup service
                         with below format

        .. code:: python

          {
               <San name>: {
                   'initiator_port_wwn_list':
                   ('200000051e55a100', '200000051e55a121'..)
                   'target_port_wwn_list':
                   ('100000051e55a100', '100000051e55a121'..)
               }
          }

    """
    target_wwns = []
    itor_tgt_map = {}
    for san_name in zone_mapping:
        one_map = zone_mapping[san_name]
        for target in one_map['target_port_wwn_list']:
            if target not in target_wwns:
                target_wwns.append(target)
        for initiator in one_map['initiator_port_wwn_list']:
            itor_tgt_map[initiator] = one_map['target_port_wwn_list']
    LOG.debug("target_wwns: %(tgt_wwns)s\n init_targ_map: %(itor_tgt_map)s",
              {'tgt_wwns': target_wwns,
               'itor_tgt_map': itor_tgt_map})
    return target_wwns, itor_tgt_map


def truncate_fc_port_wwn(wwn):
    return wwn.replace(':', '')[16:]


def is_volume_smp(volume):
    return 'smp' == extract_provider_location(volume.provider_location, 'type')


def require_consistent_group_snapshot_enabled(func):
    @six.wraps(func)
    def inner(self, *args, **kwargs):
        if not volume_utils.is_group_a_cg_snapshot_type(args[1]):
            raise NotImplementedError
        return func(self, *args, **kwargs)
    return inner


def get_remote_pool(config, volume):
    """Select remote pool name for replication.

    Prefer configured remote pool name, or same pool name
    as the source volume.
    """
    pool_name = get_pool_from_host(volume.host)
    rep_list = common.ReplicationDeviceList(config)
    remote_pool_name = rep_list[0].pool_name
    return remote_pool_name if remote_pool_name else pool_name


def is_image_cache_volume(volume):
    display_name = volume.display_name
    if (display_name.startswith('image-')
            and uuidutils.is_uuid_like(display_name[6:])):
        LOG.debug('Volume: %s is for image cache. Use sync migration and '
                  'thin provisioning.', volume.name)
        return True
    return False


def calc_migrate_and_provision(volume, default_async_migrate=True):
    """Returns a tuple of async migrate and provision type.

    The first element is the flag whether to enable async migrate,
    the second is the provision type (thin or thick).
    """
    if is_image_cache_volume(volume):
        return False, storops.VNXProvisionEnum.THIN
    else:
        specs = common.ExtraSpecs.from_volume(volume)
        return (is_async_migrate_enabled(volume, default_async_migrate),
                specs.provision)


def get_backend_qos_specs(volume):
    type_id = volume.volume_type_id
    if type_id is None:
        return None

    # Use the provided interface to avoid permission issue
    qos_specs = volume_types.get_volume_type_qos_specs(type_id)
    if qos_specs is None:
        return None

    qos_specs = qos_specs['qos_specs']
    if qos_specs is None:
        return None

    consumer = qos_specs['consumer']
    # Front end QoS specs are handled by nova. We ignore them here.
    if consumer not in common.BACKEND_QOS_CONSUMERS:
        return None

    max_iops = qos_specs['specs'].get(common.QOS_MAX_IOPS)
    max_bws = qos_specs['specs'].get(common.QOS_MAX_BWS)
    if max_iops is None and max_bws is None:
        return None

    return {
        'id': qos_specs['id'],
        common.QOS_MAX_IOPS: max_iops,
        common.QOS_MAX_BWS: max_bws,
    }
